class NoDiscordLibFound:
    def __init__(self, *args, **kwargs):
        raise ModuleNotFoundError("No discord library found. Please install a supported library.")


class FakeDiscord:
    """A fake class that is used when EzCord is used without a supported Discord library."""

    DiscordException = Exception
    CogMeta = type

    def __getattr__(self, name):
        if name in (
            "ApplicationContext",
            "AutoShardedBot",
            "Bot",
            "Cog",
            "Embed",
            "Interaction",
            "Modal",
            "Select",
            "View",
            "BadColourArgument",
        ):
            return NoDiscordLibFound

        return FakeDiscord()

    def __call__(self, *args, **kwargs):
        pass


try:
    from .dc_imports import discord

    commands = __import__(f"{discord.lib}.ext.commands", fromlist=[""])
    tasks = __import__(f"{discord.lib}.ext.tasks", fromlist=[""])

except ImportError:
    # EzCord is used without a supported Discord library
    discord = commands = FakeDiscord()  # type: ignore
    tasks, checks = None, commands  # type: ignore


if discord.__title__ == "pycord":
    from discord import AutoShardedBot, CogMeta
    from discord.ext import bridge

    slash_command = discord.slash_command
    checks = commands

else:
    CogMeta = commands.CogMeta
    AutoShardedBot = commands.AutoShardedBot
    bridge = commands

    if discord.__title__ == "discord":
        slash_command = discord.app_commands.command
        checks = discord.app_commands.checks

    elif discord.__title__ == "nextcord":
        slash_command = discord.slash_command
        checks = commands

    elif discord.__title__ == "disnake":
        slash_command = commands.slash_command
        checks = commands


discord.lib = discord.__title__  # type: ignore


PYCORD = discord.lib == "pycord"
DPY = discord.lib == "discord"



"""Embed templates that can be used to send messages.
These functions will generate embeds and send them to the desired target.

Example
-------
Here is an example for sending a success message within a Pycord application command.

.. code-block:: python

    import ezcord

    bot = ezcord.Bot()

    @bot.slash_command()
    async def hey(ctx: ezcord.EzContext):
        await ctx.success("Success!")

In any other case, the interaction must be passes to the template method.

.. code-block:: python

    class ExampleView(discord.ui.View):
        @discord.ui.button(label="Click here")
        async def button_callback(self, button, interaction):
            await emb.success(interaction, "Success!")
"""

from __future__ import annotations

import copy
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from .i18n import I18N, t
from .internal import load_embed, replace_dict, save_embeds
from .internal.dc import PYCORD, discord
from .times import convert_dt, convert_time

if PYCORD:
    _INTERACTION = (discord.Interaction, discord.ApplicationContext)
    _ctx_type = discord.ApplicationContext
else:
    _INTERACTION = (discord.Interaction,)  # type: ignore
    _ctx_type = discord.Interaction

if TYPE_CHECKING:
    import discord  # type: ignore

    _ctx_type = discord.ApplicationContext

__all__ = (
    "set_embed_templates",
    "error",
    "success",
    "warn",
    "info",
    "send",
    "EzContext",
)


def set_embed_templates(
    *,
    error_embed: discord.Embed | str | None = None,
    success_embed: discord.Embed | str | None = None,
    warn_embed: discord.Embed | str | None = None,
    info_embed: discord.Embed | str | None = None,
    **kwargs: discord.Embed,
):
    """Override the default embeds with custom ones.

    This must be called before the first embed template is used. The description
    of the embeds will be replaced with the given text.

    If you pass a string, error messages will be sent as a text instead of an embed.
    If the string is empty, the text will be taken from template methods.

    .. note::
        You can use the following variables for embed templates. They will automatically
        be replaced when the template is sent to an interaction.

        - ``{user}`` - The user who initiated the interaction
        - ``{username}`` - The name of the user
        - ``{user_mention}`` - The user mention
        - ``{user_id}`` - The ID of the user
        - ``{user_avatar}`` - The URL of the user's avatar

        Server variables will be replaced with information about the bot if the interaction
        was initiated in DMs.

        - ``{servername}`` - The guild where the interaction was initiated
        - ``{server_icon}`` - The URL of the guild's icon

        Bot variables will be replaced with information about the bot.

        - ``{guild_count}`` - Number of servers the bot is in
        - ``{user_count}`` - Number of users the bot can see
        - ``{cmd_count}`` - Number of application commands

    Parameters
    ----------
    error_embed:
        The embed to use for error messages.
    success_embed:
        The embed to use for success messages.
    warn_embed:
        The embed to use for warning messages.
    info_embed:
        The embed to use for info messages.
    **kwargs:
        Additional embed templates. Can be used with :func:`send`.

    Example
    -------
    .. code-block:: python

        from ezcord import emb

        embed = discord.Embed(
            title="Error",
            color=discord.Color.orange()
        )
        emb.set_embed_templates(error_embed=embed)
    """
    save_embeds(
        error_embed=error_embed,
        success_embed=success_embed,
        warn_embed=warn_embed,
        info_embed=info_embed,
        **kwargs,
    )


async def _send_embed(
    target: discord.Interaction | discord.abc.Messageable,
    embed: discord.Embed | str,
    ephemeral: bool = True,
    edit: bool = False,
    **kwargs,
):
    """Respond to an interaction or send an embed to a target.

    If the interaction has already been responded to,
    the message will be sent as a followup.
    """
    content = None
    if isinstance(embed, str):
        content = embed
        embed = None

    if "content" in kwargs:
        content = kwargs.pop("content")

    if not isinstance(target, _INTERACTION):
        return await target.send(content=content, embed=embed, **kwargs)

    if PYCORD and isinstance(target, discord.ApplicationContext):
        target = target.interaction

    if edit:
        if not target.response.is_done():
            return await target.response.edit_message(content=content, embed=embed, **kwargs)
        else:
            return await target.edit_original_response(content=content, embed=embed, **kwargs)

    if not target.response.is_done():
        return await target.response.send_message(
            content=content, embed=embed, ephemeral=ephemeral, **kwargs
        )
    else:
        if I18N.initialized:
            return await target.followup.send(
                content=content, embed=embed, ephemeral=ephemeral, use_locale=target, **kwargs
            )
        else:
            return await target.followup.send(
                content=content, embed=embed, ephemeral=ephemeral, **kwargs
            )


def _insert_info(
    target: discord.Interaction | discord.abc.Messageable,
    embed: discord.Embed | str,
):
    if not isinstance(target, _INTERACTION):
        return embed

    interaction = target
    if PYCORD:
        if isinstance(target, discord.ApplicationContext):
            interaction = target.interaction

    if isinstance(embed, discord.Embed):
        embed = embed.to_dict()
    embed_dic = replace_dict(embed, interaction)

    if isinstance(embed, dict):
        return discord.Embed.from_dict(embed_dic)

    return embed_dic


async def _process_message(
    target: discord.Interaction | discord.abc.Messageable,
    embed: discord.Embed | str,
    txt: str | None,
    title: str | None,
    edit: bool,
    ephemeral: bool,
    **kwargs,
):
    if isinstance(embed, discord.Embed):
        embed = copy.deepcopy(embed)
        if txt is not None:
            embed.description = txt
        if title is not None:
            embed.title = title
    elif isinstance(embed, str) and embed == "":
        embed = txt

    embed = _insert_info(target, embed)

    return await _send_embed(target, embed, ephemeral, edit, **kwargs)


def _template_docstring(params=False):
    def decorator(func, *args, **kwargs):
        func.__doc__ += """

        Parameters
        ----------
        """

        if params:
            func.__doc__ += """
        template:
            The name of the template that was used in :func:`set_embed_templates`.
        """

        func.__doc__ += """
        target:
            The target to send the message to.
        txt:
            The text for the embed description. If this is ``None``,
            you need to provide a non-empty ``Embed`` when using :func:`set_embed_templates`.
        title:
            The title of the embed. Defaults to ``None``.
        edit:
            Whether to edit the last message instead of sending a new one. Defaults to ``False``.
        ephemeral:
            Whether the message should be ephemeral. Defaults to ``True``.
        **kwargs:
            Additional keyword arguments for :meth:`discord.abc.Messageable.send`.

        Returns
        -------
        :class:`discord.Interaction` | :class:`discord.Message` | ``None``
        """
        return func

    return decorator


@_template_docstring()
async def error(
    target: discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send an error message. By default, this is a red embed."""
    embed = load_embed("error_embed")
    return await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


@_template_docstring()
async def success(
    target: discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a success message. By default, this is a green embed."""
    embed = load_embed("success_embed")
    return await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


@_template_docstring()
async def warn(
    target: discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a warning message. By default, this is a golden embed."""
    embed = load_embed("warn_embed")
    return await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


@_template_docstring()
async def info(
    target: discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send an info message. By default, this is a blue embed."""
    embed = load_embed("info_embed")
    return await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


@_template_docstring(params=True)
async def send(
    template: str,
    target: discord.Interaction | discord.abc.Messageable,
    txt: str | None = None,
    *,
    title: str | None = None,
    edit: bool = False,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a custom embed template. This needs to be set up with :func:`set_embed_templates`."""
    embed = load_embed(template)
    return await _process_message(target, embed, txt, title, edit, ephemeral, **kwargs)


class EzContext(_ctx_type):  # type: ignore
    """A custom context to access embed templates. Only works within Pycord application commands."""

    async def error(self, msg: str, **kwargs):
        return await error(self, msg, **kwargs)

    async def success(self, msg: str, **kwargs):
        return await success(self, msg, **kwargs)

    async def warn(self, msg: str, **kwargs):
        return await warn(self, msg, **kwargs)

    async def info(self, msg: str, **kwargs):
        return await info(self, msg, **kwargs)

    def t(self, key: str, count: int | None = None, **kwargs):
        return t(self.interaction, key, count, **kwargs)

    def convert_time(self, seconds: int | float, relative: bool = True):
        return convert_time(seconds, relative, use_locale=self)

    def convert_dt(self, dt: datetime | timedelta, relative: bool = True):
        return convert_dt(dt, relative, use_locale=self)



import pytest

import ezcord
from ezcord.internal.dc import FakeDiscord, commands, discord


def test_big_numbers():
    assert ezcord.format_number(0) == "0"
    assert ezcord.format_number(1) == "1"
    assert ezcord.format_number(10) == "10"
    assert ezcord.format_number(100) == "0.1K"
    assert ezcord.format_number(1_000) == "1K"
    assert ezcord.format_number(1_000_000) == "1M"
    assert ezcord.format_number(1_100_000) == "1.1M"
    assert ezcord.format_number(1_000_000_000) == "1B"

    # negative
    assert ezcord.format_number(-1) == "-1"
    assert ezcord.format_number(-100) == "-0.1K"
    assert ezcord.format_number(-1_000) == "-1K"
    assert ezcord.format_number(-1_000_000) == "-1M"
    assert ezcord.format_number(-1_100_000) == "-1.1M"
    assert ezcord.format_number(-1_000_000_000) == "-1B"

    # decimals and trailing zeros
    assert ezcord.format_number(1_550, decimal_places=2) == "1.55K"
    assert ezcord.format_number(1_550, decimal_places=3) == "1.55K"
    assert ezcord.format_number(1_550, decimal_places=3, trailing_zero=True) == "1.550K"
    assert ezcord.format_number(1_000, trailing_zero=True) == "1.0K"
    assert ezcord.format_number(1_000, decimal_places=2, trailing_zero=True) == "1.00K"


def test_convert_color():
    if isinstance(discord.lib, FakeDiscord):
        return

    assert str(ezcord.convert_color("red")) == "#e74c3c"
    assert str(ezcord.convert_color("dark red")) == "#992d22"
    assert str(ezcord.convert_color("black")) == "#000000"
    assert str(ezcord.convert_color("rgb(255, 255, 255)")) == "#ffffff"

    # test hex params
    assert str(ezcord.convert_color("ffffff")) == "#ffffff"
    assert str(ezcord.convert_color("#ffffff")) == "#ffffff"

    with pytest.raises(commands.BadColorArgument):
        assert str(ezcord.convert_color("#fff", strict_hex=True)) == "#ffffff"

    with pytest.raises(commands.BadColorArgument):
        ezcord.convert_color("ffffff", hex_hash=True)



"""
Commands to manage the bot blacklist.

This file should only be called through `bot.add_blacklist()`,
after the "blacklist" value has been set in the config.
"""

import aiosqlite

from .. import emb
from ..blacklist import _BanDB
from ..bot import Bot, Cog
from ..components import event
from ..errors import Blacklisted, ErrorMessageSent
from ..internal import EzConfig, tr
from ..internal.dc import DPY, PYCORD, commands, discord
from ..logs import log
from ..utils import create_text_file, create_yaml_file

_db = _BanDB()


async def get_or_fetch_user(bot, user_id: int):
    if user := bot.get_user(user_id):
        return user
    try:
        return await bot.fetch_user(user_id)
    except discord.NotFound:
        return None


async def _check_blacklist(interaction: discord.Interaction) -> bool:
    bans = await _db.get_bans()
    if interaction.user.id in bans and EzConfig.blacklist:
        if EzConfig.blacklist.raise_error:
            raise Blacklisted()
        else:
            await emb.error(interaction, tr("no_perms", locale=interaction))
        raise ErrorMessageSent()
    return True


@event
async def view_check(interaction: discord.Interaction):
    return await _check_blacklist(interaction)


def check_command(decorator):
    """A decorator that only applies the slash command decorator if the command is enabled."""

    def decorator_wrapper(func):
        name = func.__name__
        groups = ["blacklist", "leave"]
        for group in groups:
            name = name.replace(f"{group}_", "")

        if name not in EzConfig.blacklist.disabled_commands:
            return decorator(func)
        return func

    return decorator_wrapper


async def check_overwrite(name, *args) -> bool:
    """Returns True if the command has been overwritten, False otherwise."""
    if not EzConfig.blacklist:
        return False

    if name in EzConfig.blacklist.overwrites:
        await EzConfig.blacklist.overwrites[name](*args)
        return True

    return False


class Blacklist(Cog, hidden=True):
    def __init__(self, bot: Bot):
        super().__init__(bot)

        if DPY:
            bot.tree.interaction_check = self.global_interaction_check

    async def bot_check(self, ctx):
        """Checks if a blacklisted user is trying to use a command."""
        return await _check_blacklist(ctx)

    async def cog_check(self, ctx):
        if EzConfig.blacklist.owner_only:
            if not await self.bot.is_owner(ctx.user):
                if DPY:
                    return False
                else:
                    raise commands.CheckFailure()
        return True

    async def global_interaction_check(self, interaction: discord.Interaction):
        """Bot check for application commands in Discord.py."""
        return await _check_blacklist(interaction)

    async def interaction_check(self, interaction: discord.Interaction):
        """Cog check for application commands in Discord.py."""
        await self.cog_check(interaction)

    @Cog.listener()
    async def on_guild_join(self, guild: discord.Guild):
        bans = await _db.get_bans()
        if guild.owner and guild.owner.id in bans:
            try:
                await guild.owner.send(tr("guild_error", guild.name))
            except discord.Forbidden:
                pass
            await guild.leave()

    if PYCORD:
        admin = discord.SlashCommandGroup(
            tr("admin_group"),
            description="EzCord admin commands",
            guild_ids=EzConfig.admin_guilds,
            default_member_permissions=discord.Permissions(administrator=True),
        )
        leave = admin.create_subgroup("leave")
        blacklist = admin.create_subgroup("blacklist")
    else:
        admin = discord.app_commands.Group(
            name=tr("admin_group"),
            description="EzCord admin commands",
            guild_ids=EzConfig.admin_guilds,
            default_permissions=discord.Permissions(administrator=True),
        )
        leave = discord.app_commands.Group(
            parent=admin,
            name="leave",
            description="Make the bot leave a server",
        )
        blacklist = discord.app_commands.Group(
            parent=admin,
            name="blacklist",
            description="Manage the blacklist",
        )

    @check_command(blacklist.command(name="add", description="Add a member  blacklist"))
    # @discord.option("user", description="The user to ban/unban")
    # @discord.option("reason", description="The reason for the ban", default=None)
    async def blacklist_add(
        self,
        ctx,
        user: discord.Member,
        reason: str = None,  # type: ignore
    ):
        if await check_overwrite("add", ctx, user, reason):
            return

        if user.id == ctx.user.id:
            return await emb.error(ctx, "You can't ban yourself.")
        if user.bot:
            return await emb.error(ctx, "You can't ban a bot.")

        try:
            await _db.add_ban(user.id, reason)
        except aiosqlite.IntegrityError:
            return await emb.error(ctx, "This user is already banned.")
        await ctx.response.send_message(
            f"The user was banned successfully.\n- **Name:** {user}\n- **ID:** {user.id}",
            ephemeral=True,
        )

    @check_command(
        blacklist.command(name="remove", description="Remove a member from the blacklist")
    )
    # @discord.option("user", description="The user to ban/unban")
    async def blacklist_remove(self, ctx, user: discord.Member):
        if await check_overwrite("remove", ctx, user):
            return

        rowcount = await _db.remove_ban(user.id)
        if rowcount == 0:
            return await emb.error(ctx, "This user is not banned.")
        await ctx.response.send_message(
            f"The user **{user}** was unbanned successfully.", ephemeral=True
        )

    @check_command(blacklist.command(name="show", description="Show the bot blacklist"))
    async def blacklist_show(self, ctx):
        if await check_overwrite("show", ctx):
            return

        await ctx.response.defer(ephemeral=True)
        bans = await _db.get_full_bans()
        desc = ""

        for user_id, reason, dt in bans:
            if not reason:
                reason = "No reason provided"

            user = await get_or_fetch_user(self.bot, user_id)
            name = f"{user} - {user.id}" if user else user_id
            desc += f"[{dt.date()}] {name} - {reason}\n"

        if not desc:
            desc = "No bans found."

        file = create_text_file(desc, "bans.txt")
        await ctx.followup.send(file=file, ephemeral=True)

    @check_command(admin.command(description="Show all bot servers"))
    async def show_servers(self, ctx):
        if await check_overwrite("show_servers", ctx):
            return

        await ctx.response.defer(ephemeral=True)

        desc = [
            {
                "name": guild.name,
                "member_count": guild.member_count,
                "id": guild.id,
                "owner": f"{guild.owner} - {guild.owner.id}" if guild.owner else guild.owner_id,
                "created_at": guild.created_at,
                "joined_at": guild.me.joined_at,
                "preferred_locale": guild.preferred_locale,
                "icon": guild.icon.url if guild.icon else None,
            }
            for guild in self.bot.guilds
        ]

        file = create_yaml_file(desc, "guilds.yaml", sort_keys=False)
        await ctx.followup.send(file=file, ephemeral=True)

    @check_command(leave.command(name="server", description="Make the bot leave a server"))
    # @discord.option("guild_id", description="Leave the server with the given ID", default=None)
    async def leave_server(self, ctx, guild_id: str):
        if await check_overwrite("server", ctx, guild_id):
            return

        await ctx.response.defer(ephemeral=True)
        try:
            guild = await self.bot.fetch_guild(guild_id)
        except Exception as e:
            return await emb.error(ctx, f"I could not load this server: ```{e}```")

        await guild.leave()
        await ctx.followup.send(f"I left **{guild.name}** ({guild.id})", ephemeral=True)

    @check_command(
        leave.command(name="owner", description="Make the bot leave all guilds with a given owner")
    )
    # @discord.option("owner_id", description="Leave all servers with the specified owner")
    async def leave_owner(self, ctx, owner: discord.User):
        if await check_overwrite("owner", ctx, owner):
            return

        await ctx.defer(ephemeral=True)
        guilds = []
        member_count = 0
        for guild in self.bot.guilds:
            if guild.owner and guild.owner.id == owner.id:
                guilds.append(guild)
                member_count += guild.member_count

        return await ctx.followup.send(
            f"I found **{len(guilds)}** servers with **{owner}** as the owner "
            f"(with a total of **{member_count}** members).",
            ephemeral=True,
            view=LeaveGuilds(guilds),
        )


class LeaveGuilds(discord.ui.View):
    def __init__(self, guilds: list[discord.Guild]):
        self.guilds = guilds
        super().__init__(timeout=None)

    @discord.ui.button(label="Leave all", style=discord.ButtonStyle.red)
    async def leave(self, _: discord.ui.Button, interaction: discord.Interaction):
        if not PYCORD:
            interaction = _

        for child in self.children:
            child.disabled = True
        embed = discord.Embed(
            description="I'll now leave all servers, this may take a while."
            "\n\nI'll ping you when I'm done.",
        )
        await interaction.response.edit_message(embed=embed, view=self)

        leave_count = 0
        for guild in self.guilds:
            try:
                await guild.leave()
                leave_count += 1
            except Exception as e:
                log.warning(f"Could not leave guild **{guild.id}**: {e}")
                continue

        await interaction.followup.send(
            f"{interaction.user.mention} I successfully left **{leave_count}** servers.",
            ephemeral=True,
        )

    @discord.ui.button(label="Cancel", style=discord.ButtonStyle.grey)
    async def cancel(self, _: discord.ui.Button, interaction: discord.Interaction):
        if not PYCORD:
            interaction = _

        for child in self.children:
            child.disabled = True
        await interaction.edit(view=self)



from __future__ import annotations

import inspect
import random
import re
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Literal, Union, overload

from .internal.dc import PYCORD, discord
from .internal.deprecation import warn_deprecated
from .logs import log

MESSAGE_SEND = discord.abc.Messageable.send
MESSAGE_REPLY = discord.Message.reply
MESSAGE_EDIT = discord.Message.edit

INTERACTION_SEND = discord.InteractionResponse.send_message
INTERACTION_EDIT = discord.InteractionResponse.edit_message
INTERACTION_MODAL = discord.InteractionResponse.send_modal

WEBHOOK_SEND = discord.Webhook.send
WEBHOOK_EDIT_MESSAGE = discord.Webhook.edit_message
WEBHOOK_EDIT = discord.WebhookMessage.edit

LOCALE = Union[str]

if PYCORD:
    INTERACTION_EDIT_ORIGINAL = discord.Interaction.edit_original_response
    if hasattr(discord.Interaction, "respond"):
        # older py-cord versions might not have this method
        INTERACTION_RESPOND = discord.Interaction.respond
    else:
        INTERACTION_RESPOND = None
else:
    INTERACTION_RESPOND = None
    if hasattr(discord.Interaction, "edit_original_message"):
        INTERACTION_EDIT_ORIGINAL = discord.Interaction.edit_original_message
    else:
        INTERACTION_EDIT_ORIGINAL = None


if TYPE_CHECKING:
    import discord  # type: ignore

    LOCALE = Union[  # type: ignore
        discord.Interaction,
        discord.ApplicationContext,
        discord.InteractionResponse,
        discord.Webhook,
        discord.Guild,
        discord.Member,
        str,
    ]

__all__ = ("t", "TEmbed", "I18N", "LOCALE")


def _no_lowercase(s: str) -> bool:
    """Check if a string contains no lowercase letters."""

    return not any(c.islower() for c in s)


def t(obj: LOCALE | str, key: str, count: int | None = None, **variables) -> str:
    """Get the localized string for the given key and insert all variables.

    Parameters
    ----------
    obj:
        The object to get the locale from.
    key:
        The key of the string in the language file.
    count:
        The count for pluralization. Defaults to ``None``.
    """
    locale = I18N.get_locale(obj)
    return I18N.load_text(key, locale, count, **variables)


class TEmbed(discord.Embed):
    """A subclass of :class:`discord.Embed` for localized embeds.

    Parameters
    ----------
    key:
        The key of the embed in the language file.
    kwargs:
        Parameters from :class:`discord.Embed` or custom variables.
    """

    def __init__(self, key: str = "embed", **kwargs):
        variables, kwargs = _extract_parameters(discord.Embed.__init__, **kwargs)
        super().__init__(**kwargs)
        self.key = key
        self.variables = variables

        _, method, class_ = I18N.get_location()
        self.method_name = method
        self.class_name = class_


def _extract_parameters(func, **kwargs):
    """Extract all kwargs that are not part of the function signature and returns them as
    a dictionary of variables.
    """
    params = inspect.signature(func).parameters
    variables = {key: kwargs.pop(key) for key, value in kwargs.copy().items() if key not in params}
    return variables, kwargs


def _check_embed(locale: str, count: int | None, variables: dict, **kwargs):
    """Check if the kwargs contain an embed. Returns the updated kwargs.

    - Embed is a TEmbed: Load the embed from the language file.
    - Embed is a default Embed: Load all keys inside the embed from the language file
    """

    add_locations: tuple = ()
    embed = kwargs.get("embed")
    if isinstance(embed, TEmbed):
        variables = {**variables, **embed.variables}
        add_locations = (embed.method_name, embed.class_name)
        embed = I18N.load_embed(embed, locale)

    if embed:
        if "count" in variables:
            count = variables["count"]
        new_embed_dict = I18N.load_lang_keys(
            embed.to_dict(), locale, count, add_locations, **variables
        )
        kwargs["embed"] = discord.Embed.from_dict(new_embed_dict)

    return kwargs


def _check_embeds(locale: str, count: int | None, variables: dict, **kwargs):
    """Check if the kwargs contain an embed list. Returns the updated kwargs."""

    add_locations: tuple = ()
    embeds = kwargs.get("embeds")
    if not embeds:
        return kwargs

    new_embeds = []
    for embed in embeds:
        if isinstance(embed, TEmbed):
            variables = {**variables, **embed.variables}
            add_locations = (embed.method_name, embed.class_name)
            embed = I18N.load_embed(embed, locale)

        if "count" in variables:
            count = variables["count"]
        new_embed_dict = I18N.load_lang_keys(
            embed.to_dict(), locale, count, add_locations, **variables
        )
        new_embeds.append(discord.Embed.from_dict(new_embed_dict))

    kwargs["embeds"] = new_embeds
    return kwargs


def _check_components(component, locale: str, count: int | None, class_name: str, **variables):
    """Recursively walks through components and loads language file keys."""

    if isinstance(component, discord.ui.Container):
        for item in component.items:
            _check_components(item, locale, count, class_name, **variables)

    if isinstance(component, discord.ui.ActionRow):
        for item in component.children:
            _check_components(item, locale, count, class_name, **variables)

    if isinstance(component, discord.ui.Section):
        for item in component.items:
            _check_components(item, locale, count, class_name, **variables)

        _check_components(component.accessory, locale, count, class_name, **variables)

    # TextDisplay
    if hasattr(component, "content"):
        component.content = I18N.load_text(
            component.content, locale, count, class_name, **variables
        )

    if isinstance(component, discord.ui.Button) and hasattr(component, "label"):
        component.label = I18N.load_text(component.label, locale, count, class_name, **variables)

    if hasattr(component, "placeholder"):
        component.placeholder = I18N.load_text(
            component.placeholder, locale, count, class_name, **variables
        )

    if hasattr(component, "options"):
        for option in component.options:
            option.label = I18N.load_text(option.label, locale, count, class_name, **variables)
            option.description = I18N.load_text(
                option.description, locale, count, class_name, **variables
            )


def _check_view(locale: str, count: int | None, variables: dict, **kwargs):
    """Load all keys inside the view from the language file."""

    if "count" in variables:
        count = variables["count"]

    view = kwargs.get("view")
    if view:
        class_name = view.__class__.__name__
        for child in view.children:
            if isinstance(child, discord.ui.ActionRow):
                # check each item inside the action row with the class name of that item
                for item in child.children:
                    class_name = item.__class__.__name__
                    _check_components(item, locale, count, class_name, **variables)
                continue

            if type(child) not in [discord.ui.Select, discord.ui.Button]:
                # if a child element of the view has its own subclass, search for this class name
                # in the language file instead of the view name
                class_name = child.__class__.__name__

            _check_components(child, locale, count, class_name, **variables)

    return kwargs


def _localize_send(send_func):
    async def wrapper(
        self: (
            discord.InteractionResponse
            | discord.Webhook
            | discord.abc.Messageable
            | discord.Interaction
        ),
        content=None,
        *,
        count: int | None = None,
        locale: LOCALE | None = None,
        use_locale: LOCALE | None = None,
        **kwargs,
    ):
        """Wrapper to localize the content and the embed of a message.

        Parameters
        ----------
        self:
            The object to send the message from.
        content:
            The content of the message.
        count:
            The count for pluralization. Defaults to ``None``.
        locale:
            Use a specific object to extract the locale from. This is useful for DMs
            or followup messages. Defaults to ``None``.
        """
        if use_locale:
            warn_deprecated("use_locale", "locale", "0.7.5", "0.8.0")
        locale = locale or use_locale

        if isinstance(self, discord.Interaction):
            # This is used for cases where followup.send is executed inside of interaction.respond,
            # because the locale can't be extracted from application webhooks
            return await send_func(self, content, count=count, locale=self, **kwargs)

        locale_str = I18N.get_locale(locale or self)
        variables, kwargs = _extract_parameters(send_func, **kwargs)

        # Check content
        content = I18N.load_text(content, locale_str, **variables)

        kwargs = _check_embed(locale_str, count, variables, **kwargs)
        kwargs = _check_embeds(locale_str, count, variables, **kwargs)
        kwargs = _check_view(locale_str, count, variables, **kwargs)

        return await send_func(self, content, **kwargs)

    return wrapper


def _localize_edit(edit_func):
    async def wrapper(
        self: discord.InteractionResponse | discord.Interaction | discord.Message | discord.Webhook,
        message_id: int | None = None,
        *,
        count: int | None = None,
        locale: LOCALE | None = None,
        use_locale: LOCALE | None = None,
        **kwargs,
    ):
        """The message_id is only needed for followup.edit_message, because it's a positional
        argument in the original function.

        The parameter "locale" is only needed for followup.edit_message,
        because the locale can't be extracted automatically.
        """
        if use_locale:
            warn_deprecated("use_locale", "locale", "0.7.5", "0.8.0")
        locale = locale or use_locale

        locale_str = I18N.get_locale(locale or self)
        variables, kwargs = _extract_parameters(edit_func, **kwargs)

        # Check content (must be a kwarg)
        content = kwargs.get("content")
        if content:
            new_content = I18N.load_text(content, locale_str, count, **variables)
            kwargs["content"] = new_content

        kwargs = _check_embed(locale_str, count, variables, **kwargs)
        kwargs = _check_embeds(locale_str, count, variables, **kwargs)
        kwargs = _check_view(locale_str, count, variables, **kwargs)

        if isinstance(self, discord.Webhook):
            return await edit_func(self, message_id, **kwargs)

        return await edit_func(self, **kwargs)

    return wrapper


async def _localize_modal(
    self: discord.InteractionResponse,
    modal: discord.ui.Modal,
    *,
    count: int | None = None,
    **kwargs,
):
    locale = I18N.get_locale(self)
    variables, kwargs = _extract_parameters(INTERACTION_MODAL, **kwargs)
    modal_name = modal.__class__.__name__

    modal.title = I18N.load_text(modal.title, locale, count, modal_name, **variables)

    def check_attribute(item, attribute: str):
        if hasattr(item, attribute):
            value = getattr(item, attribute)
            localized_value = I18N.load_text(value, locale, count, modal_name, **variables)
            setattr(item, attribute, localized_value)

    for child in modal.children:
        if isinstance(child, discord.ui.Label):  # DesignerModal
            _check_components(child.item, locale, count, modal_name, **variables)

        if hasattr(child, "item"):
            check_attribute(child.item, "placeholder")

        check_attribute(child, "label")
        check_attribute(child, "description")
        check_attribute(child, "content")
        check_attribute(child, "placeholder")
        check_attribute(child, "value")

    return await INTERACTION_MODAL(self, modal)


class I18N:
    """A simple class to handle the localization of strings.

    A list of available languages is available here:
    https://discord.com/developers/docs/reference#locales

    .. note::
        Methods in this class are called automatically and do not need to be
        called manually in most cases.

    Parameters
    ----------
    localizations:
        A dictionary containing the localizations for all strings.

        If an ``en`` key is found, the values will be used for both ``en-GB`` and ``en-US``.
    fallback_locale:
        The locale to use if the user's locale is not found in the localizations.
        Defaults to ``en-US``.
    process_strings:
        Whether to replace general variables when loading the language file. Defaults to ``True``.
    prefer_user_locale:
        Whether to prefer the user's locale over the guild's locale. Defaults to ``False``.
    localize_numbers:
        This sets the thousands separator to a period or comma, depending on the current locale.
        Defaults to ``True``.
    ignore_discord_ids:
        Whether to not localize numbers that could be a Discord ID. Default to  ``True``.
        This only has an effect if ``localize_numbers`` is set to ``True``.
    do_not_localize:
        If a key starts with this string, it will not be localized. Defaults to ``".."``.
    exclude_methods:
        Method names to exclude from the search of keys in the language file.
    disable_translations:
        A list of translations to disable. Defaults to ``None``.

        The log level in :meth:`ezcord.logs.set_log` must be set to ``DEBUG`` for this to work.
    debug:
        Whether to send debug messages and warnings. Defaults to ``True``.
    language_settings:
        A function to set custom language settings.
        The function takes user/guild IDs and returns a locale or None. Defaults to ``None``.
    user_locale_settings:
        A function to dynamically set `prefer_user_locale` per guild.
        The function takes a guild IDs and returns a boolean. Defaults to ``None``.
    variables:
        Additional variables to replace in the language file. This is useful for
        values that are the same in all languages.
    """

    localizations: dict[str, dict]
    fallback_locale: str
    prefer_user_locale: bool = False
    localize_numbers: bool
    ignore_discord_ids: bool
    do_not_localize: str = ".."
    exclude_methods: list[str] | None

    _general_values: dict = {}  # general values for the current localization
    _current_general: dict = {}  # general values for the current group

    cmd_localizations: dict[str, dict] = {}  # set through bot.localize_commands
    initialized: bool = False

    _custom_language_settings: Callable[[int], int | None] | None = None
    _custom_user_locale_settings: Callable[[int], bool] | None = None

    def __init__(
        self,
        localizations: dict[str, dict],
        *,
        fallback_locale: str = "en-US",
        process_strings: bool = True,
        prefer_user_locale: bool = False,
        localize_numbers: bool = True,
        ignore_discord_ids: bool = True,
        do_not_localize: str = "..",
        exclude_methods: list[str] | None = None,
        disable_translations: (
            list[
                Literal[
                    "send",
                    "edit",
                    "reply",
                    "send_message",
                    "send_modal",
                    "edit_message",
                    "edit_original_response",
                    "webhook_send",
                    "webhook_edit_message",
                ]
            ]
            | None
        ) = None,
        debug: bool = True,
        language_settings: Callable[[int], int | None] | None = None,
        user_locale_settings: Callable[[int], bool] | None = None,
        **variables,
    ):
        I18N.initialized = True
        if "en" in localizations:
            en = localizations.pop("en")
            localizations["en-GB"] = en
            localizations["en-US"] = en

        if fallback_locale == "en":
            fallback_locale = "en-US"

        if process_strings:
            for var in variables.keys():
                if not _no_lowercase(var):
                    raise ValueError(f"Custom variable key '{var}' must be uppercase.")

            I18N.localizations = self._process_strings(localizations, **variables)
        else:
            I18N.localizations = localizations

        I18N.fallback_locale = fallback_locale
        I18N.prefer_user_locale = prefer_user_locale
        I18N.localize_numbers = localize_numbers
        I18N.ignore_discord_ids = ignore_discord_ids
        if not exclude_methods:
            exclude_methods = []
        I18N.exclude_methods = exclude_methods
        I18N.do_not_localize = do_not_localize
        I18N._custom_language_settings = language_settings
        I18N._custom_user_locale_settings = user_locale_settings

        if not disable_translations:
            disable_translations = []

        if debug:
            I18N._check_localizations()

        if "send" not in disable_translations:
            setattr(discord.abc.Messageable, "send", _localize_send(MESSAGE_SEND))
        if "edit" not in disable_translations:
            setattr(discord.Message, "edit", _localize_edit(MESSAGE_EDIT))
        if "reply" not in disable_translations:
            setattr(discord.Message, "reply", _localize_send(MESSAGE_REPLY))

        if "send_message" not in disable_translations:
            setattr(discord.InteractionResponse, "send_message", _localize_send(INTERACTION_SEND))
        if "send_modal" not in disable_translations:
            setattr(discord.InteractionResponse, "send_modal", _localize_modal)
        if "edit_message" not in disable_translations:
            setattr(discord.InteractionResponse, "edit_message", _localize_edit(INTERACTION_EDIT))
        if "edit_original_response" not in disable_translations:
            setattr(
                discord.Interaction,
                "edit_original_response",
                _localize_edit(INTERACTION_EDIT_ORIGINAL),
            )
        if "webhook_send" not in disable_translations:
            setattr(discord.Webhook, "send", _localize_send(WEBHOOK_SEND))
            setattr(discord.Interaction, "respond", _localize_send(INTERACTION_RESPOND))
        if "webhook_edit_message" not in disable_translations:
            setattr(discord.Webhook, "edit_message", _localize_edit(WEBHOOK_EDIT_MESSAGE))
        if "webhook_edit_message" not in disable_translations:
            setattr(discord.WebhookMessage, "edit_message", _localize_edit(WEBHOOK_EDIT))

    @staticmethod
    def get_locale(obj: LOCALE) -> str:
        """Get the locale from the given object. By default, this is the guild's locale.

        This method can be called even if the I18N class has not been initialized.

        Parameters
        ----------
        obj:
            The object to get the locale from.
        """

        if isinstance(obj, str):
            if hasattr(I18N, "localizations") and obj not in I18N.localizations:
                return I18N.fallback_locale
            return obj

        guild_id, user_id = None, None

        # determine interaction
        interaction, locale = None, None
        if isinstance(obj, discord.Interaction):
            interaction = obj
        elif isinstance(obj, discord.InteractionResponse):
            interaction = obj._parent
        elif isinstance(obj, discord.ApplicationContext):
            interaction = obj.interaction

        # determine locale
        elif isinstance(obj, discord.Webhook) and obj.guild:
            locale = obj.guild.preferred_locale
            guild_id = obj.guild.id
        elif isinstance(obj, discord.Member):
            locale = obj.guild.preferred_locale
            guild_id = obj.guild.id
            user_id = obj.id
        elif isinstance(obj, discord.Guild):
            locale = obj.preferred_locale
            guild_id = obj.id

        elif (
            isinstance(obj, discord.abc.Messageable | discord.Message)
            and hasattr(obj, "guild")
            and obj.guild
        ):
            locale = obj.guild.preferred_locale
            guild_id = obj.guild.id

        elif isinstance(obj, discord.User):
            # It's not possible to determine the user locale without an interaction
            locale = I18N.fallback_locale
            user_id = obj.id

        custom_locale = None

        if interaction:
            if interaction.guild and not I18N.prefer_user_locale:
                locale = interaction.guild_locale
                guild_id = interaction.guild.id
            else:
                # prevent language setting from overriding the user locale when prefer_user_locale is True
                custom_locale = interaction.locale
                user_id = interaction.user.id

            if I18N._custom_user_locale_settings and guild_id:
                prefer_user_locale = I18N._custom_user_locale_settings(guild_id)
                if prefer_user_locale:
                    custom_locale = interaction.locale
                    user_id = interaction.user.id

        # check custom language settings
        if I18N._custom_language_settings and not custom_locale:
            if guild_id:
                custom_locale = I18N._custom_language_settings(guild_id)
            if user_id:
                custom_locale = I18N._custom_language_settings(user_id)

        locale = custom_locale or locale

        # check if the locale is available. if not, use the fallback locale
        if hasattr(I18N, "localizations"):
            if locale not in I18N.localizations:
                return I18N.fallback_locale
            return locale

        return locale

    @staticmethod
    def get_clean_locale(obj: LOCALE) -> str:
        """Get the clean locale from the given object. This is the locale without the region,
        e.g. ``en`` instead of ``en-US``.

        Parameters
        ----------
        obj:
            The object to get the locale from.
        """
        locale = I18N.get_locale(obj)
        return str(locale).split("-")[0]

    @staticmethod
    def get_location():
        """Returns the name of the file, method and class for the current interaction.

        This can only get the class if a method was executed from inside the class.
        """

        inspect_stack = inspect.stack()

        # Ignore the following internal sources to determine the origin method
        # "sub" is used for regex substitution when searching for keys within other keys.
        methods = ["respond", "sub"] + I18N.exclude_methods
        files = ["i18n", "emb", "interactions"]

        file, method, class_ = None, None, None
        for i in inspect_stack[2:]:
            if i.function not in methods and Path(i.filename).stem not in files:
                try:
                    class_ = i.frame.f_locals["self"].__class__.__name__
                except KeyError:
                    pass  # No class found
                file = i.filename
                method = i.function
                break

        if file:
            file = Path(file).stem

        return file, method, class_

    @staticmethod
    def _replace_variables(string: str, locale: str, **variables):
        """Replace all given variables in the string. Supports localized numbers.

        Example
        -------
        >>> I18N._replace_variables("Hello {name}", name="Timo")
        "Hello Timo"
        """
        if not string:
            return string

        for key, value in variables.items():
            if I18N.localize_numbers and isinstance(value, int):
                if not (I18N.ignore_discord_ids and len(str(value)) >= 17):
                    value = f"{value:,}"
                    if locale == "de":
                        value = value.replace(",", ".")
            string = string.replace("{" + key + "}", str(value))

        return string

    @staticmethod
    def _get_text(
        key: str, locale: str, count: int | None, called_class: str | None, add_locations: tuple
    ) -> str:
        """Looks for the specified key in different locations of the language file."""

        if key.startswith(I18N.do_not_localize):
            return key.replace(I18N.do_not_localize, "", 1)

        file_name, method_name, class_name = I18N.get_location()

        lookups: list[list | tuple]
        if "." in key:
            lookups = [key.split("."), [file_name] + key.split(".")]
        else:
            lookups = [
                (file_name, method_name, key),
                (file_name, called_class, key),
                (file_name, class_name, key),
                (file_name, "general", key),
                ("general", key),
                (file_name, key),
            ]
            for location in add_locations:
                lookups.append((file_name, location, key))

        localizations = I18N.localizations[locale]

        for lookup in lookups:
            current_section = localizations.copy()
            for location in lookup:
                try:
                    current_section = current_section.get(location, {})
                except AttributeError:
                    return key

            txt = current_section
            if isinstance(txt, str):
                return txt
            elif isinstance(txt, int):
                return str(txt)
            elif isinstance(txt, list):
                return random.choice(txt)
            elif count is not None and isinstance(txt, dict):
                # Load pluralization if available
                if count == 0 and "zero" in txt:
                    return txt["zero"]
                elif count == 1 and "one" in txt:
                    return txt["one"]
                elif count > 1 and "many" in txt:
                    return txt["many"]
                elif str(count) in txt:
                    return txt[str(count)]

        return key

    @overload
    @staticmethod
    def load_text(
        key: None,
        locale: str,
        count: int | None = ...,
        called_class: str | None = ...,
        add_locations: tuple = ...,
        **variables,
    ) -> None: ...

    @overload
    @staticmethod
    def load_text(
        key: str,
        locale: str,
        count: int | None = ...,
        called_class: str | None = ...,
        add_locations: tuple = ...,
        **variables,
    ) -> str: ...

    @staticmethod
    def load_text(
        key: str,
        locale: str,
        count: int | None = None,
        called_class: str | None = None,
        add_locations: tuple = (),
        **variables,
    ) -> str | None:
        """A helper methods that calls :meth:`get_text` to load the specified key
        and :meth:`replace_variables` to replace the variables.

        A class name can be given if the kwargs contain a view or modal.
        This name will be used to load strings from the init method or from decorators, as the
        class can only be fetched automatically if a method was executed from inside the class.

        Additional locations can be added to the lookup by passing a tuple of strings.
        This is used to load embed keys from the location of the embed creation,
        instead of the location of the embed usage.
        """

        if key is None:
            return None

        string = I18N._get_text(key, locale, count, called_class, add_locations)

        if count:
            variables = {**variables, "count": count}
        string = I18N._replace_variables(string, locale, **variables)

        def replace_keys(m: re.Match):
            k = m.group(1)
            check_key = I18N._get_text(k, locale, count, called_class, add_locations)
            return check_key if check_key != k else m.group()

        # check if key contains other keys
        if "{" in string and "}" in string:
            string = re.sub(r"{(.*?)}", replace_keys, string)

        return I18N._replace_variables(string, locale, **variables)

    @staticmethod
    def load_embed(embed: TEmbed, locale: str) -> discord.Embed:
        """Loads an embed from the language file."""

        file_name, method_name, class_name = I18N.get_location()

        # search not only the location of the embed usage,
        # but also the location of the embed creation
        original_method, original_class = embed.method_name, embed.class_name

        lookups: list[list | tuple]
        if "." in embed.key:
            lookups = [embed.key.split("."), [file_name] + embed.key.split(".")]
        else:
            lookups = [
                (file_name, method_name, embed.key),
                (file_name, original_method, embed.key),
                (file_name, original_class, embed.key),
                (file_name, class_name, embed.key),
                (file_name, embed.key),
            ]

        localizations = I18N.localizations[locale]

        for lookup in lookups:
            current_section = localizations.copy()
            for location in lookup:
                current_section = current_section.get(location, {})

            if current_section:
                t_embed_dict = embed.to_dict()
                for key, value in current_section.items():
                    t_embed_dict[key] = value

                return discord.Embed.from_dict(t_embed_dict)

        return discord.Embed(description=embed.key, color=discord.Color.blurple())

    @staticmethod
    def load_lang_keys(
        content: dict | str,
        locale: str,
        count: int | None = None,
        add_locations: tuple = (),
        **variables,
    ) -> dict | str:
        """Iterates through the content, loads the keys from the language file
        and replaces all variables with their values.

        Does not modify the original content.
        """

        if isinstance(content, str):
            return I18N.load_text(content, locale, count, add_locations=add_locations, **variables)

        content = deepcopy(content)

        for key, value in content.items():
            if isinstance(value, str):
                if key in ["color", "colour", "type", "url", "timestamp", "image", "thumbnail"]:
                    continue

                content[key] = I18N.load_text(
                    value, locale, count, add_locations=add_locations, **variables
                )
            elif isinstance(value, list):
                items = []
                for element in value:
                    items.append(
                        I18N.load_lang_keys(element, locale, count, add_locations, **variables)
                    )
                content[key] = items
            elif isinstance(value, dict):
                content[key] = I18N.load_lang_keys(value, locale, count, add_locations, **variables)

        return content

    @staticmethod
    def _replace_general_variables(string: str) -> str:
        """Replaces global and local general variables with their values."""

        def replace_local(match: re.Match):
            match = match.group().replace("{.", "").replace("}", "")
            if match in I18N._current_general:
                return I18N._current_general[match]

            if match in I18N._general_values:
                return I18N._general_values[match]

            return match

        def replace_global(possible_match: re.Match) -> str:
            match = possible_match.group()
            clean_match = match.replace("{", "").replace("}", "")
            if clean_match in I18N._general_values and _no_lowercase(str(clean_match)):
                if type(I18N._general_values[clean_match]) is str:
                    return I18N._general_values[clean_match]

            return str(match)

        string = re.sub(r"{\..*?}", replace_local, string)
        string = re.sub(r"{.*?}", replace_global, string)
        return string

    @staticmethod
    def _replace_dict(content: dict | str) -> dict | str:
        """Iterates through the content and replaces all general variables with their values.

        This is only needed once when loading the language file.
        """
        if isinstance(content, dict) and "general" in content:
            I18N._current_general = content["general"]

        if isinstance(content, str):
            return I18N._replace_general_variables(content)

        for key, value in content.items():
            if isinstance(value, str):
                content[key] = I18N._replace_general_variables(value)
            if isinstance(value, list):
                items = []
                for element in value:
                    items.append(I18N._replace_dict(element))
                content[key] = items
            elif isinstance(value, dict):
                content[key] = I18N._replace_dict(value)

        return content

    @staticmethod
    def _process_strings(localizations: dict, **variables) -> dict:
        """Process all strings and replace general variables when loading the language file.

        A general variable is defined in one of the "general" sections of the language files.
        """
        new_dict = {}
        for locale, values in localizations.items():
            if "general" in values:
                I18N._general_values = {**values["general"], **variables}
            else:
                I18N._general_values = variables

            new_dict[locale] = I18N._replace_dict(values)

        return new_dict

    @staticmethod
    def _find_missing_keys(fallback: dict, current_locale: dict):
        """Find keys and sub-keys that are missing in the current locale."""

        missing_keys = []

        def explore_dict(original: dict, current: dict, path: str):
            for key, value in original.items():
                if key not in current:
                    missing_keys.append(f"{path}.{key}".lstrip("."))
                elif isinstance(value, dict) and isinstance(current.get(key), dict):
                    explore_dict(value, current[key], f"{path}.{key}")

        explore_dict(fallback, current_locale, "")
        return missing_keys

    @staticmethod
    def _check_localizations():
        """Checks if all locales have the same keys."""

        for locale, values in I18N.localizations.items():
            missing_keys = I18N._find_missing_keys(I18N.localizations[I18N.fallback_locale], values)
            if len(missing_keys) > 0:
                log.warning(
                    f"Locale '{locale}' misses some keys from the fallback locale: {missing_keys}"
                )



__title__ = "ezcord"
__author__ = "tibue99"
__license__ = "MIT"
__version__ = "0.7.5"

from . import blacklist, components, emb, i18n, logs, sql, times, utils
from .bot import *
from .components import *
from .emb import *
from .enums import *
from .errors import *
from .i18n import *
from .logs import *
from .sql import *
from .times import *
from .utils import *



from __future__ import annotations

import json
import os
from functools import cache
from pathlib import Path

from ...internal import discord
from ...logs import log
from ..config import EzConfig


@cache
def load_lang(language: str) -> dict[str, dict[str, str]]:
    """Loads the given language file and checks if the user provided a custom language file."""

    if language == "auto":
        language = EzConfig.default_lang

    lang = {}
    parent = Path(__file__).parent.absolute()
    for element in os.scandir(parent):
        if not element.is_file() or not element.name.endswith(f"{language}.json"):
            continue

        with open(os.path.join(parent, f"{language}.json"), encoding="utf-8") as file:
            lang = json.load(file)
            break

    # check if the user has a custom language file
    for root, directories, files in os.walk(os.getcwd()):
        for filename in files:
            if filename != f"ez_{language}.json":
                continue

            log.debug(f"Custom language file loaded: **{filename}**")

            path = os.path.join(root, filename)
            with open(path, encoding="utf-8") as user_file:
                user_dic = json.load(user_file)
                for category, values in user_dic.items():
                    for value in values:
                        if category not in lang:
                            lang[category] = {}
                        lang[category][value] = values[value]

    if EzConfig.lang != "auto" and lang == {}:
        log.warning(f"Language file for language '{language}' not found. Falling back to 'en'.")

    return lang


def localize_command(
    command: discord.ApplicationCommand | discord.SlashCommandGroup,
    locale: str,
    localizations: dict,
    default_locale: str,
):
    """Localize a slash command or a slash command group.

    This code is inspired by https://github.com/Dorukyum/pycord-i18n,
    which is licensed under the MIT License.
    """

    if isinstance(command, discord.SlashCommandGroup):
        for cmd in command.walk_commands():
            localize_command(cmd, locale, localizations.get(cmd.name, {}), default_locale)

    if name := localizations.get("name"):
        if locale == default_locale:
            command.name = name

        if command.name_localizations is discord.MISSING:
            command.name_localizations = {locale: name}
        else:
            command.name_localizations[locale] = name

    if not isinstance(command, discord.SlashCommand):
        return

    if description := localizations.get("description"):
        if locale == default_locale:
            command.description = description

        if command.description_localizations is discord.MISSING:
            command.description_localizations = {locale: description}
        else:
            command.description_localizations[locale] = description

    if options := localizations.get("options"):
        for option_name, localization in options.items():
            if option := discord.utils.get(command.options, name=option_name):
                if option_name := localization.get("name"):
                    if locale == default_locale:
                        option.name = option_name

                    if option.name_localizations is discord.MISSING:
                        option.name_localizations = {locale: option_name}
                    else:
                        option.name_localizations[locale] = option_name

                if option_desc := localization.get("description"):
                    if locale == default_locale:
                        option.description = option_desc

                    if option.description_localizations is discord.MISSING:
                        option.description_localizations = {locale: option_desc}
                    else:
                        option.description_localizations[locale] = option_desc

                if option_choices := localization.get("choices"):
                    for choice in option.choices:
                        if isinstance(choice, str):
                            choice = discord.OptionChoice(name=choice)

                        name = option_choices.get(choice.name)
                        if name:
                            if locale == default_locale:
                                choice.name = name

                            if choice.name_localizations is discord.MISSING:
                                choice.name_localizations = {locale: name}
                            else:
                                choice.name_localizations[locale] = name


def localize_cog(
    cog_name: str,
    cog,
    locale: str,
    localizations: dict,
):
    """Localize a cog. This is only used for the help command."""

    if localized_cog := localizations.get(cog_name):
        if localized_name := localized_cog.get("name"):
            if not hasattr(cog, "name_localizations"):
                cog.name_localizations = {locale: localized_name}
            else:
                cog.name_localizations[locale] = localized_name

        if localized_desc := localized_cog.get("description"):
            if not hasattr(cog, "description_localizations"):
                cog.description_localizations = {locale: localized_desc}
            else:
                cog.description_localizations[locale] = localized_desc



"""Classes that are adding some functionality for default components.

.. hint::

        Components in your code are automatically replaced with the ones from this module.
        You can use `discord.ui.View` instead of `ezcord.View`.
"""

from __future__ import annotations

import asyncio
import inspect
from typing import Callable

from .internal.dc import discord
from .internal.deprecation import warn_deprecated
from .logs import log

_view_checks: list[Callable] = []
_view_check_failures: list[Callable] = []

__all__ = ("event", "Modal", "View", "EzView", "EzModal", "DropdownPaginator")


def _check_coro(func):
    if not asyncio.iscoroutinefunction(func):
        raise TypeError(f"Event registered must be a coroutine function, not {type(func)}")


def _check_params(func, amount):
    parameters = inspect.signature(func).parameters
    if len(parameters) != amount:
        raise ValueError(
            f"Event method must have '{amount}' parameters, has '{len(parameters)}' instead"
        )


def event(coro):
    """A decorator to register custom checks and error handlers for Discord components.

    Example
    -------
    .. code-block:: python

        @ezcord.event
        async def view_check(interaction):
            return interaction.user.id == 12345  # Returns True or False

        @ezcord.event
        async def on_view_check_failure(interaction):
            await interaction.response.send_message("You can't use this!")
    """
    _check_coro(coro)

    name = coro.__name__.lstrip("_")

    if name == "view_check":
        _check_params(coro, 1)
        _view_checks.append(coro)
    elif name == "on_view_check_failure":
        _check_params(coro, 1)
        _view_check_failures.append(coro)
    elif name == "on_view_error":
        raise ValueError(
            f"This event was removed in version 0.7.5, use 'Bot.on_view_error' instead: "
            f"https://docs.pycord.dev/en/master/api/clients.html#discord.Bot.on_modal_error"
        )
    elif name == "on_modal_error":
        raise ValueError(
            f"This event was removed in version 0.7.5, use 'Bot.on_modal_error' instead: "
            f"https://docs.pycord.dev/en/master/api/clients.html#discord.Bot.on_view_error"
        )
    else:
        raise ValueError(f"Invalid event name: '{coro.__name__}'")

    log.debug(f"Registered event **{coro.__name__}**")


class View(discord.ui.View):
    ignore_timeout_errors: bool = False

    """Base class for all :class:`discord.ui.View` classes that adds some functionality.

    Parameters
    ----------
    ignore_timeout_error:
        If ``True``, views will not raise an exception if an error occurs during the timeout event.
    """

    def __init__(self, *args, ignore_timeout_error: bool = False, **kwargs):
        self.ignore_timeout_error = ignore_timeout_error
        super().__init__(*args, **kwargs)

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        """Returns ``True`` if all custom checks return ``True`` or if no custom checks are registered."""
        for coro in _view_checks:
            if not await coro(interaction):
                return False
        return True

    async def on_check_failure(self, interaction: discord.Interaction) -> None:
        """This method is called if :meth:`interaction_check` returns ``False``."""
        for coro in _view_check_failures:
            await coro(interaction)

    async def on_timeout(self) -> None:
        """If ``disable_on_timeout`` is set to ``True``, this will disable all components,
        unless the viw has been explicitly stopped.
        """
        try:
            return await super().on_timeout()
        except discord.NotFound:
            return
        except discord.HTTPException as e:
            if self.ignore_timeout_error or View.ignore_timeout_errors:
                return
            log.exception(
                f"Error in View **{type(self).__name__}** ({type(self).__module__}) ```{e}```",
                exc_info=e,
            )


class Modal(discord.ui.Modal):
    """Base class for all :class:`discord.ui.Modal` classes."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class EzView(View):
    """Alias for :class:`View`."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        warn_deprecated("ezcord.EzView", "discord.ui.View", "0.7", "0.8")


class EzModal(Modal):
    """Alias for :class:`Modal`."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        warn_deprecated("ezcord.EzModal", "discord.ui.Modal", "0.7", "0.8")


# replace all default components with Ezcord components
discord.ui.View = View
discord.ui.Modal = Modal


class DropdownPaginator(discord.ui.Select):
    """A dropdown paginator that can be used to paginate through a list of options.

    This is useful when we have more options than Discord allows in a single dropdown menu.

    Parameters
    ----------
    options:
        The options that will be displayed in the dropdown.
    next_page_label:
        The label of the next page button.
    previous_page_label:
        The label of the previous page button.
    next_page_emoji:
        The emoji of the next page button.
    previous_page_emoji:
        The emoji of the previous page button.
    page:
        The current page of the dropdown.
    **kwargs:
        Additional keyword arguments that are passed to :class:`discord.ui.Select`.
    """

    def __init__(
        self,
        options: list[discord.SelectOption],
        next_page_label: str = "Next page",
        previous_page_label: str = "Previous page",
        next_page_emoji: str = "⬇️",
        previous_page_emoji: str = "⬆️",
        page: int = 0,
        **kwargs,
    ):
        self.page = page

        self.next_page_label = next_page_label
        self.previous_page_label = previous_page_label
        self.next_page_emoji = next_page_emoji
        self.previous_page_emoji = previous_page_emoji

        self.total_options = options
        self.current_options = self.load_options(options, self.page)
        self.kwargs = kwargs

        super().__init__(options=self.current_options, **kwargs)

    @property
    def item_selected(self):
        """Returns ``True`` if the user selected an item from the dropdown.
        If the user clicked on the next or previous page button, this will return ``False``.
        """

        return "ez_next" not in self.values and "ez_previous" not in self.values

    async def callback(self, interaction: discord.Interaction):
        """Edit the dropdown menu if the user selects a page option."""

        def set_current_options(x):
            if isinstance(x, DropdownPaginator):
                self.options = self.current_options
                return self
            else:
                return x

        if self.check_next_page():
            new_children = map(set_current_options, self.view.children)

            self.view.children = list(new_children)
            await interaction.response.edit_message(view=self.view)
            return

        elif self.check_previous_page():
            new_children = map(set_current_options, self.view.children)

            self.view.children = list(new_children)
            await interaction.response.edit_message(view=self.view)
            return

    def check_next_page(self) -> bool:
        """Returns True if the user clicked on the next page button.
        In this case, the dropdown menu will be edited.
        """
        if "ez_next" in self.values:
            self.page += 1
            self.current_options = self.load_options(self.total_options, self.page)
            return True
        return False

    def check_previous_page(self) -> bool:
        """Returns True if the user clicked on the previous page button.
        In this case, the dropdown menu will be edited.
        """
        if "ez_previous" in self.values:
            self.page -= 1
            self.current_options = self.load_options(self.total_options, self.page)
            return True
        return False

    def load_options(
        self, options: list[discord.SelectOption], chunk: int = 0
    ) -> list[discord.SelectOption]:
        """Split the options into chunks and append options for next/previous pages."""

        chunk_size = 23
        if len(options) > chunk_size:
            x = [
                options[option : option + chunk_size]
                for option in range(0, len(options), chunk_size)
            ]

            new_options = x[chunk]
            if len(new_options) == chunk_size:
                new_options.append(
                    discord.SelectOption(
                        label=self.next_page_label, value="ez_next", emoji=self.next_page_emoji
                    )
                )

            if chunk > 0:
                new_options.insert(
                    0,
                    discord.SelectOption(
                        label=self.previous_page_label,
                        value="ez_previous",
                        emoji=self.previous_page_emoji,
                    ),
                )
        else:
            new_options = options

        return new_options



from .internal.dc import DPY, discord

if DPY:
    _DC_ERROR_TYPE = discord.app_commands.AppCommandError
else:
    _DC_ERROR_TYPE = discord.DiscordException


class ErrorMessageSent(_DC_ERROR_TYPE):  # type: ignore
    """Exception that can be raised to indicate that an error message has already been sent to the user.

    This could be useful if an error message has already been sent to the user within a check
    function. This prevents further messages to the user, as this error will be ignored by
    the default error handler.
    """


class EzcordException(Exception):
    """Base exception class for all Ezcord exceptions."""


class ConvertTimeError(EzcordException):
    """Raised when a time conversion fails."""


class DurationError(ConvertTimeError):
    """Raised when a given duration is too long."""


class Blacklisted(EzcordException):
    """Can be raised when a blacklisted user tries to use a command.

    This error can be caught in a command error handler to send a custom response.
    """


class InvalidFormat(EzcordException):
    """Raised when an invalid URL has been passed."""


class MessageNotFound(EzcordException):
    """Raised when a message is not found in a certain channel."""


class ChannelNotFound(EzcordException):
    """Raised when a channel is not found."""


class GuildMismatch(MessageNotFound):
    """Raised when a message does not belong to the specified guild."""


class MissingPermission(EzcordException):
    """Raised when the bot does not have permission to access a certain object."""



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Literal


@dataclass
class Blacklist:
    db_path: str
    db_name: str
    raise_error: bool
    owner_only: bool
    disabled_commands: list[EzConfig.BLACKLIST_COMMANDS]
    overwrites: dict[str, Callable]


class EzConfig:
    """A class to store configuration values.

    These values are usually set only once, but are used throughout the runtime of the bot.
    """

    BLACKLIST_COMMANDS = Literal["add", "remove", "show", "owner", "server", "show_servers"]

    lang: str = "en"
    default_lang: str = "en"
    embed_templates: dict = {}

    # Blacklist
    admin_guilds: list[int] | None = None
    blacklist: Blacklist | None = None



from __future__ import annotations

import json
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

import asyncpg


def _process_args(args) -> tuple:
    """If SQL query parameters are passed as a tuple instead of single values,
    the tuple will be unpacked.
    """

    if len(args) == 1 and isinstance(args, tuple):
        if isinstance(args[0], tuple):
            args = args[0]

    # convert dict to str
    return tuple(json.dumps(arg) if isinstance(arg, dict) else arg for arg in args)


def _process_one_result(row: asyncpg.Record, default: Any):
    row = row[0] if row is not None and len(row) == 1 else row
    return row if row is not None else default


def _process_exec_status(status: str) -> QueryStatus:
    status_list = status.split(" ")
    if len(status_list) > 1 and not status_list[1].isdigit():
        status_list[0] += " " + status_list[1]
        status_list.pop(1)

    query_type = status_list[0]
    if len(status_list) == 2:
        return QueryStatus(type=query_type, rowcount=int(status_list[1]))
    elif len(status_list) == 3:
        return QueryStatus(
            type=query_type, rowcount=int(status_list[1]), inserts=int(status_list[2])
        )
    else:
        return QueryStatus(type=query_type)


@dataclass
class QueryStatus:
    """A class to access the status of a :meth:`PGHandler.exec` call."""

    type: str
    rowcount: int = 0
    inserts: int = 0


class EzConnection(asyncpg.Connection):
    """A subclass of :class:`asyncpg.Connection` that adds aliases
    to be compatible with the sqlite handler.
    """

    async def one(self, sql: str, *args, default=None, **kwargs):
        row = await super().fetchrow(sql, *_process_args(args), **kwargs)
        return _process_one_result(row, default)

    async def all(self, sql: str, *args, **kwargs) -> list:
        result = await super().fetch(sql, *_process_args(args), **kwargs)
        if result and len(result[0]) == 1:
            return [row[0] for row in result]
        return result

    async def fetchval(self, sql: str, *args, default=None, **kwargs):
        value = await super().fetchval(sql, *_process_args(args), **kwargs)
        return value or default

    async def exec(self, sql: str, *args, **kwargs) -> QueryStatus:
        status = await super().execute(sql, *_process_args(args), **kwargs)
        return _process_exec_status(status)

    async def execute(self, *args, **kwargs) -> QueryStatus:
        """Alias for :meth:`exec`."""
        return await self.exec(*args, **kwargs)


class PGHandler:
    """A class that provides helper methods for PostgreSQL databases.

    .. note::

            It's recommended to set the database connection parameters in the ``.env`` file.

            - Reference: https://www.postgresql.org/docs/current/libpq-envars.html

    Parameters
    ----------
    custom_pool:
        Override the default connection pool with a key. Each custom pool has a unique key.
        Defaults to ``None``.
    auto_setup:
        Whether to call :meth:`setup` when the first instance of this class is created.
        Defaults to ``True``.
    **kwargs:
        Keyword arguments for :func:`asyncpg.create_pool`.
    """

    pool: asyncpg.Pool | None = None
    _pools: dict[str, asyncpg.Pool | None] = {}

    _auto_setup: list[PGHandler] = []
    _auto_pool: list[PGHandler] = []

    def __init__(
        self,
        *,
        custom_pool: str | None = None,
        auto_setup: bool = True,
        **kwargs,
    ):
        self.kwargs = kwargs
        self.custom_pool = custom_pool

        if auto_setup and self not in self._auto_setup:
            PGHandler._auto_setup.append(self)

        if custom_pool:
            if self not in self._auto_pool:
                PGHandler._auto_pool.append(self)
            if custom_pool not in PGHandler._pools:
                PGHandler._pools[custom_pool] = None

    async def _check_pool(self) -> asyncpg.Pool:
        """Create a new connection pool or returns an existing one.

        Custom pools are stored in :attr:`_pools`. If a custom pool for a specified key already
        exists, it will be returned and set as the pool for the current class instance.
        """

        if self.custom_pool:
            if self._pools[self.custom_pool]:
                self.pool = self._pools[self.custom_pool]
                return self._pools[self.custom_pool]
        elif PGHandler.pool is not None:
            return PGHandler.pool

        pool = await asyncpg.create_pool(connection_class=EzConnection, **self.kwargs)

        if self.custom_pool:
            PGHandler._pools[self.custom_pool] = pool
            self.pool = pool
            return self.pool
        else:
            PGHandler.pool = pool
            return PGHandler.pool

    async def one(self, sql: str, *args, default=None, **kwargs):
        """Returns one result record. If no record is found, ``None`` is returned.

        Parameters
        ----------
        sql:
            The SQL query to execute.
        *args:
            Arguments for the query.
        default:
            When the query returns no results, this value will be returned instead of ``None``.

        Returns
        -------
        The result record or ``None``.
        """
        pool = await self._check_pool()

        async with pool.acquire() as con:
            return await con.one(sql, *args, default=default, **kwargs)

    async def all(self, sql: str, *args, **kwargs) -> list:
        """Returns all result records.

        Parameters
        ----------
        sql:
            The SQL query to execute.
        *args:
            Arguments for the query.

        Returns
        -------
        A list of result records.
        """
        pool = await self._check_pool()

        async with pool.acquire() as con:
            return await con.all(sql, *args, **kwargs)

    async def fetchval(self, sql: str, *args, default=None, **kwargs):
        """Returns one value.

        Parameters
        ----------
        sql:
            The SQL query to execute.
        *args:
            Arguments for the query.
        default:
            When the query returns no results, this value will be returned instead of ``None``.

        Returns
        -------
        The value or ``None``.
        """
        pool = await self._check_pool()

        async with pool.acquire() as con:
            return await con.fetchval(sql, *args, default=default, **kwargs)

    async def exec(self, sql: str, *args, **kwargs) -> QueryStatus:
        """Executes a SQL query.

        Parameters
        ----------
        sql:
            The SQL query to execute.
        *args:
            Arguments for the query.
        """
        pool = await self._check_pool()

        async with pool.acquire() as con:
            return await con.exec(sql, *args, **kwargs)

    async def execute(self, sql: str, *args, **kwargs) -> QueryStatus:
        """Alias for :meth:`exec`."""

        return await self.exec(sql, *args, **kwargs)

    async def executemany(self, sql: str, args: Iterable[Iterable[Any]], **kwargs) -> str:
        """Executes a SQL multiquery.

        Parameters
        ----------
        sql:
            The multiquery to execute.
        *args:
            Arguments for the multiquery.
        """
        pool = await self._check_pool()

        async with pool.acquire() as con:
            return await con.executemany(sql, args, **kwargs)



# This is used to import PGHandler while asyncpg not installed. Once the class is initialized
# by the user, an error is thrown.


class PGHandler:
    _auto_setup: list = []

    def __init__(self, *args, **kwargs):
        raise ModuleNotFoundError(
            "Please install the 'asyncpg' package to use the PostgreSQL handler."
        )



"""Inspired by Pycord: https://github.com/Pycord-Development/pycord/blob/master/discord/utils.py."""

from __future__ import annotations

import warnings


def warn_deprecated(
    name: str,
    instead: str | None = None,
    since: str | None = None,
    removed: str | None = None,
    reference: str | None = None,
    stacklevel: int = 3,
) -> None:
    """Warn about a deprecated function, with the ability to specify details about the deprecation. Emits a
    DeprecationWarning.

    Parameters
    ----------
    name: str
        The name of the deprecated function.
    instead:
        A recommended alternative to the function.
    since:
        The version in which the function was deprecated.
    removed:
        The version in which the function is planned to be removed.
    reference:
        A reference that explains the deprecation, typically a URL to a page such as a changelog entry or a GitHub
        issue/PR.
    stacklevel:
        The stacklevel kwarg passed to :func:`warnings.warn`. Defaults to ``3``.
    """
    warnings.simplefilter("always", DeprecationWarning)
    message = f"'{name}' is deprecated"
    if since:
        message += f" since version {since}"
    if removed:
        message += f" and will be removed in version {removed}"
    if instead:
        message += f", consider using '{instead}' instead"
    message += "."
    if reference:
        message += f" See {reference} for more information."

    warnings.warn(message, stacklevel=stacklevel, category=DeprecationWarning)
    warnings.simplefilter("default", DeprecationWarning)



from .colors import *
from .config import EzConfig
from .embed_templates import *
from .language.languages import load_lang, localize_cog, localize_command
from .ready_style import *
from .translation import *



from __future__ import annotations

import inspect
import random
from copy import deepcopy

from .. import emb
from ..bot import Bot, Cog
from ..components import View
from ..enums import HelpStyle
from ..i18n import I18N
from ..internal import fill_custom_variables, replace_embed_values, tr
from ..internal.dc import PYCORD, discord, slash_command
from ..logs import log


def get_emoji(cog: Cog) -> str:
    if hasattr(cog, "emoji") and cog.emoji:
        emoji = cog.emoji
    else:
        emoji = random.choice(["🔰", "👻", "🍪", "👥", "🦕", "🐧", "✨", "😩", "🍔"])

    return emoji


def get_group(cog: Cog, cog_name: str, locale: str) -> tuple[str | None, str]:
    group = None
    if hasattr(cog, "group") and cog.group:
        group = cog.group

    name = group if group else cog_name

    localized_name = None
    if hasattr(cog, "name_localizations"):
        localized_name = cog.name_localizations.get(locale, cog_name)

    try:
        # if the command has a group, the localized name will be loaded from the group cog instead
        localized_name = I18N.cmd_localizations[locale]["cogs"][name]["name"]
    except (KeyError, AttributeError):
        pass

    name = localized_name or name
    return group, name


def replace_placeholders(s: str, **kwargs: str):
    for key, value in kwargs.items():
        if not value:
            continue
        s = s.replace(f"{{{key}}}", value)
    return s


def get_cmd_desc(command, locale: str):
    if command.description_localizations is not discord.MISSING:
        return command.description_localizations.get(locale, command.description)
    return command.description


def get_cog_desc(cog, locale: str) -> str | None:
    if hasattr(cog, "description_localizations"):
        return cog.description_localizations.get(locale, cog.description)
    return None


def get_perm_parent(cmd: discord.SlashCommand) -> discord.SlashCommandGroup | None:
    """Iterates through parent groups until it finds a group with default_member_permissions set."""
    if PYCORD:
        while cmd.default_member_permissions is None:
            cmd = cmd.parent
            if cmd is None:
                return None

        return cmd.default_member_permissions

    else:
        while cmd.default_permissions is None:
            cmd = cmd.parent
            if cmd is None:
                return None

        return cmd.default_permissions


async def pass_checks(command: discord.SlashCommand, ctx) -> bool:
    """Returns True if the current user passes all checks (Pycord only)."""
    passed = True
    for check in deepcopy(command.checks):
        try:
            if inspect.iscoroutinefunction(check):
                await check(ctx)
            else:
                if not check(ctx):
                    passed = False
                    break
        except Exception:
            passed = False
            break

    return passed


class Help(Cog, hidden=True):
    def __init__(self, bot: Bot):
        super().__init__(bot)
        if PYCORD:
            if bot.help.contexts:
                self.help.contexts = bot.help.contexts
            if bot.help.integration_types:
                self.help.integration_types = bot.help.integration_types
        else:
            self.help.guild_only = bot.help.guild_only

    @slash_command(name=tr("cmd_name"), description=tr("cmd_description"))
    async def help(self, ctx):
        interaction = ctx.interaction if PYCORD else ctx
        embed = self.bot.help.embed
        if embed is None:
            embed = discord.Embed(
                title=tr("embed_title", locale=ctx.interaction), color=discord.Color.blue()
            )

        # check language file for embed localization
        locale = I18N.get_locale(ctx)
        try:
            embed_overrides = I18N.localizations[locale]["help"]["embed"]
        except (KeyError, AttributeError):
            # KeyError: language file for this locale does not have a help/embed section
            # AttributeError: I18N class is not in use
            embed_overrides = {}

        for key, value in embed_overrides.items():
            setattr(embed, key, value)

        embed = replace_embed_values(
            embed, interaction, await fill_custom_variables(self.bot.help.kwargs)
        )

        options = []
        commands: dict[str, dict] = {}
        for name, cog in self.bot.cogs.items():
            if hasattr(cog, "hidden") and cog.hidden:
                continue

            group, name = get_group(cog, name, locale)

            if len(name) == 0:
                log.warning(
                    "A cog has a name with length 0. "
                    "This cog will not be displayed in the help command."
                )
                continue

            if len(name) > 100:
                name = name[:90] + "..."

            if name not in commands:
                commands[name] = {}
            if "cmds" not in commands[name]:
                commands[name]["cmds"] = []

            if len(cog.get_commands()) == 0 and PYCORD:
                continue

            emoji = get_emoji(cog)
            commands[name]["emoji"] = emoji

            desc = get_cog_desc(cog, locale)
            if not desc:
                desc = tr("default_description", name, locale=ctx)
                if not desc:
                    log.warning(
                        f"The default description for cog '{name}' is invalid. "
                        f"This can be changed in the language file."
                    )

            if "description" not in commands[name]:
                commands[name]["description"] = desc

            field_name = replace_placeholders(self.bot.help.title, name=name, emoji=emoji)
            desc = replace_placeholders(
                self.bot.help.description, description=desc, name=name, emoji=emoji
            )

            if PYCORD:
                cog_cmds = [
                    cmd
                    for cmd in cog.walk_commands()
                    if isinstance(cmd, discord.ApplicationCommand)
                    and type(cmd)
                    not in [
                        discord.MessageCommand,
                        discord.UserCommand,
                        discord.SlashCommandGroup,
                    ]
                ]
            else:
                cog_cmds = cog.walk_app_commands()

            for command in cog_cmds:
                if PYCORD:
                    default_perms = command.default_member_permissions
                    guild_ids = command.guild_ids
                else:
                    default_perms = command.default_permissions
                    guild_ids = command._guild_ids

                if self.bot.help.permission_check:
                    if PYCORD and not await pass_checks(command, ctx):
                        continue

                    if isinstance(ctx.user, discord.Member):
                        if default_perms and not command.parent:
                            if not default_perms.is_subset(ctx.user.guild_permissions):
                                continue

                        parent_perms = get_perm_parent(command)
                        if parent_perms and not parent_perms.is_subset(ctx.user.guild_permissions):
                            continue

                if ctx.guild and guild_ids and ctx.guild.id not in guild_ids:
                    continue

                commands[name]["cmds"].append(command)

            cmd_count = len(commands[name]["cmds"])

            if cmd_count == 0:
                continue
            if not group:
                if self.bot.help.show_cmd_count:
                    label = f"{name} ({cmd_count})"
                else:
                    label = name
                option = discord.SelectOption(label=label, emoji=emoji, value=name)
                options.append(option)
                if self.bot.help.show_categories:
                    embed.add_field(name=field_name, value=desc, inline=False)

        if len(options) == 0:
            return await ctx.response.send_message(tr("no_commands", locale=ctx), ephemeral=True)
        if len(options) > 25 or len(embed.fields) > 25:
            log.error(
                f"Help command category limit reached. Only 25 out of {len(options)} are shown."
            )
            options = options[:25]
            embed.fields = embed.fields[:25]

        sorted_options = sorted(options, key=lambda x: [char for char in x.label if char.isalpha()])
        embed.fields = sorted(embed.fields, key=lambda x: x.name.lower())
        view = CategoryView(sorted_options, self.bot, ctx.user, commands, ctx)
        for button in self.bot.help.buttons:
            view.add_item(deepcopy(button))
        await ctx.response.send_message(view=view, embed=embed, ephemeral=self.bot.help.ephemeral)


class CategorySelect(discord.ui.Select):
    def __init__(
        self,
        options: list[discord.SelectOption],
        bot: Bot,
        member: discord.Member | discord.User,
        commands: dict[str, dict],
        interaction,
    ):
        super().__init__(
            min_values=1,
            max_values=1,
            placeholder=tr("placeholder", locale=interaction),
            options=options,
        )
        self.bot = bot
        self.member = member
        self.commands = commands

    def get_mention(self, cmd, locale: str) -> str:
        """This is only needed for Discord.py."""
        if self.bot.all_dpy_commands:
            for c in self.bot.all_dpy_commands:
                if c.name == cmd.name:
                    cmd = c
                    break

        if cmd.name_localizations is not discord.MISSING:
            default = cmd.name_localizations.get(locale, f"**/{cmd.qualified_name}**")
        else:
            default = f"**/{cmd.qualified_name}**"

        try:
            return cmd.mention or default
        except AttributeError:
            return default

    async def callback(self, interaction: discord.Interaction):
        if self.bot.help.author_only and interaction.user != self.member:
            return await emb.error(interaction, tr("wrong_user", locale=interaction))

        locale = I18N.get_locale(interaction)
        title = self.values[0]
        cmds = self.commands[title]
        emoji = cmds["emoji"]

        embed = self.bot.help.embed
        if embed is None:
            embed = discord.Embed(
                color=discord.Color.blue(),
            )
        else:
            embed = replace_embed_values(
                embed, interaction, await fill_custom_variables(self.bot.help.kwargs)
            )
        embed.title = replace_placeholders(self.bot.help.title, name=title, emoji=emoji)
        embed.clear_fields()

        if self.bot.help.show_description:
            embed.description = desc = cmds["description"] + "\n"
        else:
            desc = ""

        commands = cmds["cmds"]
        embed_field_styles = [
            HelpStyle.embed_fields,
            HelpStyle.codeblocks_inline,
            HelpStyle.codeblocks,
        ]
        style = self.bot.help.style
        if len(commands) > 25 and style in embed_field_styles:
            style = HelpStyle.embed_description

        if style == HelpStyle.embed_fields:
            for command in commands:
                embed.add_field(
                    name=f"**{self.get_mention(command, locale)}**",
                    value=f"`{get_cmd_desc(command, locale)}`",
                    inline=False,
                )

        elif style == HelpStyle.codeblocks or style == HelpStyle.codeblocks_inline:
            for command in commands:
                embed.add_field(
                    name=f"**{self.get_mention(command, locale)}**",
                    value=f"```{get_cmd_desc(command, locale)}```",
                    inline=style == HelpStyle.codeblocks_inline,
                )

        elif style == HelpStyle.embed_description:
            embed.description = desc + "\n"
            for command in commands:
                if len(embed.description) <= 3500:
                    embed.description += f"**{self.get_mention(command, locale)}**\n{get_cmd_desc(command, locale)}\n\n"
                else:
                    log.error("Help embed length limit reached. Some commands are not shown.")
                    break

        elif style == HelpStyle.markdown:
            embed.description = desc
            for command in commands:
                if len(embed.description) <= 3500:
                    embed.description += f"### {self.get_mention(command, locale)}\n{get_cmd_desc(command, locale)}\n"
                else:
                    log.error("Help embed length limit reached. Some commands are not shown.")
                    break

        if len(commands) == 0:
            embed.description = tr("no_commands", locale=interaction)

        view = CategoryView(self.options, self.bot, self.member, self.commands, interaction)
        for button in self.bot.help.buttons:
            view.add_item(deepcopy(button))
        await interaction.response.edit_message(embed=embed, view=view)


class CategoryView(View):
    def __init__(
        self,
        options: list[discord.SelectOption],
        bot: Bot,
        member: discord.Member | discord.User,
        commands: dict[str, dict],
        interaction: discord.Interaction,
    ):
        if PYCORD:
            super().__init__(timeout=bot.help.timeout, disable_on_timeout=True)
        else:
            super().__init__(timeout=None)

        self.add_item(CategorySelect(options, bot, member, commands, interaction))



from __future__ import annotations

import json
from collections.abc import Iterable
from copy import deepcopy
from typing import Any

import aiosqlite


class DBHandler:
    """A class that provides helper methods for SQLite databases.

    Parameters
    ----------
    path:
        The path to the database file.
    connection:
        A connection to the database. If not provided, a new connection will be created.
        If ``auto_connect`` is ``True``, this will be ignored.
    auto_connect:
        Automatically create a new connection that will be used for all queries.
        This is used by :meth:`start`.

        When used without a context manager, this must be closed with :meth:`close`
        or by using ``end=True`` in :meth:`exec`.
    auto_setup:
        Whether to call :meth:`setup` when the first instance of this class is created. Defaults to ``True``.
        This is called in the ``on_ready`` event of the bot.
    conv_json:
        Whether to auto-convert JSON. Defaults to ``False``.
    foreign_keys:
        Whether to enforce foreign keys. Defaults to ``False``.
    **kwargs:
        Keyword arguments for :func:`aiosqlite.connect`.

    Example
    -------
    You can use this class with an asynchronous context manager:

    .. code-block:: python3

            async with DBHandler("ezcord.db") as db:
                await db.exec("CREATE TABLE IF NOT EXISTS vip (id INTEGER PRIMARY KEY, name TEXT)")
                await db.exec("INSERT INTO vip (name) VALUES (?)", "Timo")
    """

    _auto_setup: list[DBHandler] = []

    def __init__(
        self,
        path: str,
        *,
        connection: aiosqlite.Connection | None = None,
        auto_connect: bool = False,
        auto_setup: bool = True,
        conv_json: bool = False,
        foreign_keys: bool = False,
        **kwargs,
    ):
        self.DB = path
        self.connection = connection
        self.auto_connect = auto_connect
        self.conv_json = conv_json
        self.foreign_keys = foreign_keys
        self.kwargs = kwargs

        if auto_setup and self not in self._auto_setup:
            DBHandler._auto_setup.append(self)

    async def __aenter__(self):
        self.auto_connect = True
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        return await self.close()

    def _process_args(self, args) -> tuple:
        """If SQL query parameters are passed as a tuple instead of single values,
        the tuple will be unpacked.

        If ``conv_json`` is ``True``, all dicts will be converted to JSON strings.
        """

        if len(args) == 1 and isinstance(args, tuple):
            if isinstance(args[0], tuple):
                args = args[0]

        if self.conv_json:
            json_args: tuple = ()
            for arg in args:
                if isinstance(arg, dict):
                    json_args = json_args + (json.dumps(arg),)
                else:
                    json_args = json_args + (arg,)
            args = json_args

        return args

    def start(
        self, *, conv_json: bool | None = None, foreign_keys: bool | None = None, **kwargs
    ) -> DBHandler:
        """Opens a new connection with the current DB settings. Additional settings can
        be provided as keyword arguments.

        This should be used as an asynchronous context manager. The connection will commit
        automatically after exiting the context manager.

        Parameters
        ----------
        conv_json:
            Whether to auto-convert JSON.
        foreign_keys:
            Whether to enforce foreign keys.
        **kwargs:
            Additional keyword arguments for :func:`aiosqlite.connect`.

        Example
        -------
        .. code-block:: python3

            class VipDB(DBHandler):
                def __init__(self):
                    super().__init__("ezcord.db")

                async def setup(self):
                    async with self.start() as db:
                        await db.exec(
                            "CREATE TABLE IF NOT EXISTS vip (id INTEGER PRIMARY KEY, name TEXT)"
                        )
                        await db.exec("INSERT INTO vip (name) VALUES (?)", "Timo")
        """
        cls = deepcopy(self)
        cls.auto_connect = True
        cls.kwargs = {**self.kwargs, **kwargs}

        # override settings if provided
        if conv_json is not None:
            cls.conv_json = conv_json
        if foreign_keys is not None:
            cls.foreign_keys = foreign_keys

        return cls

    async def connect(self, **kwargs) -> DBHandler:
        """Alias for :meth:`start`."""

        return self.start(**kwargs)

    async def _connect(self, **kwargs) -> aiosqlite.Connection:
        """Connect to an SQLite database. If the class instance has an active connection,
        that connection will be returned instead.

        If ``auto_connect`` is ``True``, a connection will be created and stored
        as the instance connection.
        """

        if self.connection is not None:
            return self.connection

        if "conv_json" in kwargs:
            self.conv_json = kwargs.pop("conv_json")
        if "foreign_keys" in kwargs:
            self.foreign_keys = kwargs.pop("foreign_keys")

        con_args = {**kwargs, **self.kwargs}
        con = await aiosqlite.connect(self.DB, **con_args)

        if self.auto_connect:
            self.connection = con

        if self.foreign_keys:
            await con.execute("PRAGMA foreign_keys = ON")

        return con

    async def close(self):
        """Commits and closes the current connection to the database.

        This is called automatically when using :meth:`start` as a context manager.
        """
        if self.connection is not None:
            await self.connection.commit()
            await self.connection.close()

    async def _close(self, db: aiosqlite.Connection):
        if not self.connection:
            await db.close()

    async def _commit_and_close(self, db: aiosqlite.Connection, end: bool = False):
        if not self.connection or end:
            await db.commit()
            await db.close()

    @staticmethod
    def _convert_tuple_json(result: tuple) -> tuple:
        if not isinstance(result, tuple):
            result = (result,)

        new_result: tuple = ()
        for value in result:
            if isinstance(value, str):
                try:
                    value = json.loads(value)
                except json.JSONDecodeError:
                    pass
            new_result = new_result + (value,)

        return new_result

    def _convert_json_one(self, result: tuple) -> tuple:
        """Converts all JSON strings from :meth:`one` to dicts (if enabled)."""

        if not self.conv_json or not result:
            return result

        return self._convert_tuple_json(result)

    def _convert_json_all(self, result: list) -> list:
        """Converts all JSON strings from :meth:`all` to dicts (if enabled)."""

        if self.conv_json:
            return [self._convert_tuple_json(row) for row in result]
        return result

    async def one(self, sql: str, *args, fill: bool = False, **kwargs):
        """Returns one result row. If no row is found, ``None`` is returned.

        If the query returns only one column, the value of that column is returned.

        Parameters
        ----------
        sql:
            The SQL query to execute.
        *args:
            Arguments for the query.
        fill:
            Whether to return ``None`` for all selected values if no row is found.
            Defaults to ``False``.
        **kwargs:
            Keyword arguments for the connection.

        Returns
        -------
        The result row or ``None``. A result row is either a tuple or a single value.
        """
        args = self._process_args(args)
        db = await self._connect(**kwargs)
        try:
            async with db.execute(sql, args) as cursor:
                result = await cursor.fetchone()
        except Exception as e:
            await self._close(db)
            raise e

        await self._close(db)

        if result is None:
            if fill:
                return (None,) * len(cursor.description)
            return None

        result = self._convert_json_one(result)
        if len(result) == 1:
            return result[0]

        return result

    async def all(self, sql: str, *args, **kwargs) -> list:
        """Returns all result rows.

        If the query returns only one column, the values of that column are returned.

        Parameters
        ----------
        sql:
            The SQL query to execute.
        *args:
            Arguments for the query.
        **kwargs:
            Keyword arguments for the connection.

        Returns
        -------
        A list of result rows. A result row is either a tuple or a single value.
        """
        args = self._process_args(args)
        db = await self._connect(**kwargs)
        try:
            async with db.execute(sql, args) as cursor:
                result = await cursor.fetchall()
        except Exception as e:
            await self._close(db)
            raise e

        await self._close(db)

        result = self._convert_json_all(result)
        if len(result) == 0 or len(result[0]) == 1:
            return [row[0] for row in result]

        return result

    async def exec(self, sql: str, *args, end: bool = False, **kwargs) -> aiosqlite.Cursor:
        """Executes a SQL query.

        Parameters
        ----------
        sql:
            The SQL query to execute.
        end:
            Whether to commit and close the connection after executing the query.
        *args:
            Arguments for the query.
        **kwargs:
            Keyword arguments for the connection.
        """
        args = self._process_args(args)
        db = await self._connect(**kwargs)
        try:
            cursor = await db.execute(sql, args)
        except Exception as e:
            await self._commit_and_close(db, end)
            raise e
        await self._commit_and_close(db, end)
        return cursor

    async def execute(self, sql: str, *args, end: bool = False, **kwargs) -> aiosqlite.Cursor:
        """Alias for :meth:`exec`."""
        return await self.exec(sql, *args, end=end, **kwargs)

    async def executemany(
        self, sql: str, args: Iterable[Iterable[Any]], **kwargs
    ) -> aiosqlite.Cursor:
        """Executes a SQL multiquery.

        Parameters
        ----------
        sql:
            The mutliquery to execute.
        *args:
            Arguments for the mutliquery.
        **kwargs:
            Keyword arguments for the connection.
        """
        db = await self._connect(**kwargs)
        try:
            cursor = await db.executemany(sql, args)
        except Exception as e:
            await self._commit_and_close(db)
            raise e
        await self._commit_and_close(db)
        return cursor



"""Some utility functions for Discord and Python."""

from __future__ import annotations

import asyncio
import fnmatch
import inspect
import io
import itertools
import json
import os
import random
import re
import warnings
from pathlib import Path
from typing import Any

import yaml

from .errors import (
    ChannelNotFound,
    GuildMismatch,
    InvalidFormat,
    MessageNotFound,
    MissingPermission,
)
from .internal import get_locale
from .internal.dc import DPY, commands, discord

__all__ = (
    "create_json_file",
    "create_text_file",
    "create_html_file",
    "create_yaml_file",
    "avatar",
    "random_avatar",
    "codeblock",
    "ez_autocomplete",
    "count_lines",
    "load_message",
    "format_number",
    "convert_color",
)


def create_json_file(
    dictionary: dict,
    filename: str = "data.json",
    *,
    indent: int = 2,
    description: str | None = None,
    spoiler: bool = False,
    **kwargs,
) -> discord.File:
    """Create a :class:`discord.File` object from a dictionary.

    Parameters
    ----------
    dictionary:
        The dictionary to convert to a JSON file.
    filename:
        The filename to use for the JSON file.
    indent:
        The indent to use for the JSON file.
    description:
        The description to use for the discord file.
    spoiler:
        Whether the Discord file should be a spoiler.
    **kwargs:
        Additional keyword arguments for :py:func:`json.dumps`.

    Returns
    -------
    :class:`discord.File`
    """
    content = json.dumps(dictionary, indent=indent, **kwargs).encode()
    return discord.File(
        io.BytesIO(content),
        filename=filename,
        description=description,
        spoiler=spoiler,
    )


def create_text_file(text: str, filename: str = "data.txt", **kwargs) -> discord.File:
    """Create a :class:`discord.File` object from a string.

    Parameters
    ----------
    text:
        The string to convert to a text file.
    filename:
        The filename to use for the text file.
    **kwargs:
        Keyword arguments for :class:`discord.File`.

    Returns
    -------
    :class:`discord.File`
    """
    return discord.File(io.BytesIO(text.encode()), filename=filename, **kwargs)


def create_html_file(html: str, filename: str = "data.html", **kwargs) -> discord.File:
    """Create a :class:`discord.File` object from an HTML string.

    Parameters
    ----------
    html:
        The HTML string to convert to an HTML file.
    filename:
        The filename to use for the HTML file.
    **kwargs:
        Keyword arguments for :class:`discord.File`.

    Returns
    -------
    :class:`discord.File`
    """
    return create_text_file(html, filename, **kwargs)


def create_yaml_file(
    data: dict | list,
    filename: str = "data.yaml",
    *,
    indent: int = 2,
    description: str | None = None,
    spoiler: bool = False,
    **kwargs,
) -> discord.File:
    """Create a :class:`discord.File` object from a YAML string.

    Parameters
    ----------
    data:
        The data to convert to a YAML file.
    filename:
        The filename to use for the YAML file.
    indent:
        The indent to use for the YAML file.
    description:
        The description to use for the discord file.
    spoiler:
        Whether the Discord file should be a spoiler.
    **kwargs:
        Keyword arguments for :meth:`yaml.dump`.

    Returns
    -------
    :class:`discord.File`
    """

    yaml_string = yaml.dump(data, indent=indent, **kwargs)
    return discord.File(
        io.BytesIO(yaml_string.encode()),
        filename=filename,
        description=description,
        spoiler=spoiler,
    )


def avatar(user_id: int) -> str:
    """Returns a permanent URL of a user's avatar.

    Parameters
    ----------
    user_id:
        The ID of the user.
    """
    return f"https://api.cookieapp.me/avatar/{user_id}"


def random_avatar() -> str:
    """Returns the URL of a random default avatar."""

    return f"https://cdn.discordapp.com/embed/avatars/{random.randint(0, 5)}.png"


def codeblock(
    content: int | str,
    *,
    lang: str = "yaml",
    unit: str = "",
    interaction: discord.Interaction | None = None,
) -> str:
    """Returns a codeblock with the given content.

    Parameters
    ----------
    content:
        The content of the codeblock. If the content is an integer, it will be
        formatted with commas (or dots if the language is German).
    lang:
        The language of the codeblock. Defaults to ``yaml``.
    unit:
        The text to display after the given content. This is only used if the content is an integer.
    interaction:
        The interaction to get the language from. Defaults to ``None``.
        If not provided, the language will be set to the default language.
        The language will determine how large numbers are formatted.
    """

    if isinstance(content, int):
        number = f"{content:,}"
        if get_locale(interaction) == "de":
            number = number.replace(",", ".")
        block = f"```{lang}\n{number}"
        if unit:
            block += f" {unit}"
        block += "```"
        return block

    return f"```{lang}\n{content}```"


def ez_autocomplete(values):
    """A similar function as :func:`~discord.utils.basic_autocomplete`, but instead of returning
    options starting with the user's value, it returns options containing the user's value.

    Parameters
    ----------
    values: Iterable[str]
        Accepts an iterable of :class:`str`, a callable (sync or async) that takes a single argument
        of :class:`~discord.AutocompleteContext`, or a coroutine. Must resolve to an iterable of :class:`str`.
    """

    async def autocomplete_callback(ctx):
        _values = values

        if callable(_values):
            _values = _values(ctx)
        if asyncio.iscoroutine(_values):
            _values = await _values

        def check(item: Any) -> bool:
            item = getattr(item, "name", item)
            return str(ctx.value or "").lower() in str(item).lower()

        gen = (val for val in _values if check(val))
        return iter(itertools.islice(gen, 25))

    return autocomplete_callback


def count_lines(
    directory: str | None = None,
    *,
    count_empty_lines: bool = True,
    include_hidden: bool = False,
    ignored_dirs: list[str] | None = None,
    ignored_files: list[str] | None = None,
) -> int:
    """Counts the total amount of lines in all Python files in the given directory.

    Parameters
    ----------
    directory:
        The directory to count the lines in. Defaults to the current working directory.
    count_empty_lines:
        Whether to count empty lines. Defaults to ``True``.
    include_hidden:
        Whether to include directories starting with a dot. Defaults to ``False``.
    ignored_dirs:
        A list of directories to ignore. By default, venv folders and folders starting with a dot
        are ignored.
    ignored_files:
        A list of file patterns to ignore.
    """
    if directory is None:
        directory = os.getcwd()
    if ignored_dirs is None:
        ignored_dirs = []
    if ignored_files is None:
        ignored_files = []

    total_lines = 0
    for root, _, files in os.walk(directory):
        if not include_hidden and Path(root).parts[-1].startswith("."):
            ignored_dirs.append(root)
        if "pyvenv.cfg" in files:  # ignore venv folders
            ignored_dirs.append(root)

        if any([True for pattern in ignored_dirs if pattern in str(Path(root))]):
            continue

        for file in files:
            if not file.endswith(".py"):
                continue

            if any([True for pat in ignored_files if fnmatch.fnmatch(file, pat)]):
                continue

            file_path = os.path.join(root, file)
            with open(file_path, errors="ignore") as f:
                for line in f:
                    if not count_empty_lines and line.strip() == "":
                        continue
                    total_lines += 1

    return total_lines


async def load_message(
    obj: discord.Guild | discord.Bot, message_url: str, error: bool = False
) -> discord.Message | None:
    """Get a message from a message URL.

    Parameters
    ----------
    obj:
        The object to use to fetch the message.
    message_url:
        The URL of the message.
    error:
        Whether to raise an error if the message couldn't be found. If this is ``False``,
        the function will return ``None`` if the message couldn't be found. Defaults to ``False``.

    Raises
    ------
    InvalidFormat
        The message URL is invalid.
    MissingPermission
        The bot does not have permissions to access the channel.
    ChannelNotFound
        The channel was not found.
    MessageNotFound
        The message couldn't be found in the given channel, perhaps it was deleted.
    GuildMismatch
        The message does not belong to the given guild.
    """
    pattern = r"channels\/(\d+)\/(\d+)\/(\d+)"  # /channels/guild_id/channel_id/message_id
    matches = re.search(pattern, message_url)

    if not matches or len(matches.groups()) != 3:
        if error:
            raise InvalidFormat("The message URL is invalid.")
        return None

    _, channel_id, message_id = matches.groups()

    try:
        channel = await discord.utils.get_or_fetch(obj, "channel", channel_id)
    except discord.Forbidden:
        if error:
            raise MissingPermission("The bot does not have permissions to access the channel.")
        return None
    except discord.NotFound:
        if error:
            raise ChannelNotFound("The channel was not found.")
        return None
    except discord.InvalidData:
        if error:
            raise GuildMismatch("The message does not belong to the given guild.")
        return None

    try:
        message = await channel.fetch_message(message_id)
    except discord.NotFound:
        if error:
            raise MessageNotFound(
                "The message couldn't be found in the given channel, perhaps it was deleted."
            )
        return None

    return message


def format_number(number: int, *, decimal_places: int = 1, trailing_zero: bool = False) -> str:
    """Format a big number to short and readable format.

    Parameters
    ----------
    number:
        The number to format.
    decimal_places:
        The amount of decimal places to display. Defaults to ``1``.
    trailing_zero:
        Whether to show trailing zeros. Defaults to ``False``.

    Returns
    -------
    :class:`str`
        The formatted number.

    Example
    -------
    >>> format_number(1_000_000)
    '1M'
    >>> format_number(1_550, decimal_places=2)
    '1.55K'
    >>> format_number(1_000, trailing_zero=True)
    '1.0K'
    """

    suffix = ""
    if number >= 1_000_000_000 or number <= -1_000_000_000:
        txt = f"{number / 1_000_000_000:.{decimal_places}f}"
        suffix = "B"
    elif number >= 1_000_000 or number <= -1_000_000:
        txt = f"{number / 1_000_000:.{decimal_places}f}"
        suffix = "M"
    elif number >= 100 or number <= -100:
        txt = f"{number / 1_000:.{decimal_places}f}"
        suffix = "K"
    else:
        txt = str(number)

    if not trailing_zero and "." in txt:
        txt = txt.rstrip("0").rstrip(".")

    return txt + suffix


def convert_color(color: str, strict_hex: bool = True, hex_hash: bool = False) -> discord.Colour:
    """Convert a color string to a :class:`discord.Color`.

    Parameters
    ----------
    color:
        The color to convert. This can be a hex code, a color name, or an RGB value.
    strict_hex:
        Whether hex codes must have a length of 6. Defaults to ``True``.
    hex_hash:
        Whether a hex code must start with a hash. Defaults to ``False``.

    Returns
    -------
    :class:`discord.Color`
        The converted color.

    Raises
    ------
    :exc:`commands.BadColourArgument`
        The color could not be converted.
    """

    additional_colors = {
        "white": "#FFFFFF",
        "black": "#000000",
        "grey": "dark_grey",
    }
    if not DPY:
        additional_colors["pink"] = "#eb459f"  # Pycord only has 'nitro_pink'

    for key, value in additional_colors.items():
        if color == key:
            color = value

    if DPY:
        conv = discord.colour
    else:
        conv = commands.ColorConverter()

    if color[0:2] == "0x":
        rest = color[2:]
        if rest.startswith("#"):
            return conv.parse_hex_number(rest[1:])
        return conv.parse_hex_number(rest)

    arg = color.lower()
    if arg[0:3] == "rgb":
        return conv.parse_rgb(arg)

    arg = arg.replace(" ", "_")
    method = getattr(discord.Colour, arg, None)
    if not (arg.startswith("from_") or method is None or not inspect.ismethod(method)):
        return method()

    if hex_hash and not color.startswith("#"):
        raise commands.BadColourArgument(color)

    maybe_hex = color.lstrip("#")
    if (strict_hex and len(maybe_hex) != 6) or any(c not in "0123456789abcdef" for c in maybe_hex):
        raise commands.BadColourArgument(color)

    return conv.parse_hex_number(color.lstrip("#"))



import os
import tempfile

import pytest

import ezcord
from ezcord.internal.dc import discord


class UserDB(ezcord.DBHandler):
    def __init__(self, path):
        super().__init__(path)

    async def setup(self):
        await self.exec(
            """CREATE TABLE IF NOT EXISTS users(
            user_id INTEGER PRIMARY KEY,
            coins INTEGER DEFAULT 0
            )"""
        )

    async def add_coins(self, user_id, amount):
        async with self.start() as cursor:
            await cursor.exec("INSERT OR IGNORE INTO users (user_id) VALUES (?)", (user_id,))
            await cursor.exec(
                "UPDATE users SET coins = coins + ? WHERE user_id = ?",
                (amount, user_id),
            )

    async def get_users(self):
        return await self.all("SELECT user_id FROM users")

    async def get_coins(self, user_id):
        return await self.one("SELECT coins FROM users WHERE user_id = ?", (user_id,))


@pytest.mark.dc
def test_libs():
    """Test compatibility with different Discord libraries."""

    intents = discord.Intents.default()
    intents.message_content = True
    bot = ezcord.Bot(command_prefix="!", intents=intents)

    assert bot
    with pytest.raises(TypeError):
        bot.run()


@pytest.mark.asyncio
async def test_db():
    """Test the DBHandler class."""

    with tempfile.NamedTemporaryFile(delete=False) as db_file:
        db = UserDB(db_file.name)

        await db.setup()
        await db.add_coins(12345, 100)

        assert await db.get_coins(12345) == 100

        await db.add_coins(54321, 0)
        assert await db.get_users() == [12345, 54321]

    if db_file.name:
        os.remove(db_file.name)


@pytest.mark.dc
def test_i18n():
    class InvalidType:
        pass

    invalid_type = InvalidType()

    if discord.__name__ == "pycord":  # Only Pycord is supported by i18n
        assert ezcord.I18N.get_locale("en") == "en"
        assert ezcord.I18N.get_locale("en-US") == "en-US"
        assert ezcord.I18N.get_locale(invalid_type) is None

        assert ezcord.I18N.get_clean_locale("en-US") == "en"



try:
    from .postgresql import PGHandler
except ImportError:
    # Import fake class that throws an error on initialization if asyncpg is not installed.
    try:
        from .fake_pg import PGHandler  # type: ignore
    except ImportError:
        # This will never be executed, but is needed for type hints, TYPE_CHECKING failed me.
        from .postgresql import PGHandler

from .sqlite import DBHandler
